#------------------------------------------------------------------------------
#
# Copyright (c) Microsoft Corporation.
# All rights reserved.
#
# This code is licensed under the MIT License.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files(the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions :
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#------------------------------------------------------------------------------

from .application import (
    ClientApplication,
    ConfidentialClientApplication,
    PublicClientApplication,
    )
from .oauth2cli.assertion import AutoRefresher
from .oauth2cli.oidc import Prompt, IdTokenError
from .sku import __version__
from .token_cache import TokenCache, SerializableTokenCache
from .auth_scheme import PopAuthScheme
from .managed_identity import (
    SystemAssignedManagedIdentity, UserAssignedManagedIdentity,
    ManagedIdentityClient,
    ManagedIdentityError,
    ArcPlatformNotSupportedError,
    )

# Putting module-level exceptions into the package namespace, to make them
# 1. officially part of the MSAL public API, and
# 2. can still be caught by the user code even if we change the module structure.
from .oauth2cli.oauth2 import BrowserInteractionTimeoutError




# Copyright (c) Microsoft Corporation.
# All rights reserved.
#
# This code is licensed under the MIT License.
import copy
import hashlib
import hmac
import json
import logging
import os
import ssl
import sys
import time
import uuid
from urllib.parse import urlparse  # Python 3+
from collections import UserDict  # Python 3+
from typing import List, Optional, Union  # Needed in Python 3.7 & 3.8
import requests
from requests.adapters import HTTPAdapter
from urllib3.connection import HTTPSConnection
from urllib3.connectionpool import HTTPSConnectionPool
from .token_cache import TokenCache
from .individual_cache import _IndividualCache as IndividualCache
from .throttled_http_client import ThrottledHttpClientBase, RetryAfterParser
from .cloudshell import _is_running_in_cloud_shell
from .sku import SKU, __version__


logger = logging.getLogger(__name__)


class ManagedIdentityError(ValueError):
    pass


class ManagedIdentity(UserDict):
    """Feed an instance of this class to :class:`msal.ManagedIdentityClient`
    to acquire token for the specified managed identity.
    """
    # The key names used in config dict
    ID_TYPE = "ManagedIdentityIdType"  # Contains keyword ManagedIdentity so its json equivalent will be more readable
    ID = "Id"

    # Valid values for key ID_TYPE
    CLIENT_ID = "ClientId"
    RESOURCE_ID = "ResourceId"
    OBJECT_ID = "ObjectId"
    SYSTEM_ASSIGNED = "SystemAssigned"

    _types_mapping = {  # Maps type name in configuration to type name on wire
        CLIENT_ID: "client_id",
        RESOURCE_ID: "msi_res_id",  # VM's IMDS prefers msi_res_id https://github.com/Azure/azure-rest-api-specs/blob/dba6ed1f03bda88ac6884c0a883246446cc72495/specification/imds/data-plane/Microsoft.InstanceMetadataService/stable/2018-10-01/imds.json#L233-L239
        OBJECT_ID: "object_id",
    }

    @classmethod
    def is_managed_identity(cls, unknown):
        return (isinstance(unknown, ManagedIdentity)
            or cls.is_system_assigned(unknown)
            or cls.is_user_assigned(unknown))

    @classmethod
    def is_system_assigned(cls, unknown):
        return isinstance(unknown, SystemAssignedManagedIdentity) or (
            isinstance(unknown, dict)
            and unknown.get(cls.ID_TYPE) == cls.SYSTEM_ASSIGNED)

    @classmethod
    def is_user_assigned(cls, unknown):
        return isinstance(unknown, UserAssignedManagedIdentity) or (
            isinstance(unknown, dict)
            and unknown.get(cls.ID_TYPE) in cls._types_mapping
            and unknown.get(cls.ID))

    def __init__(self, identifier=None, id_type=None):
        # Undocumented. Use subclasses instead.
        super(ManagedIdentity, self).__init__({
            self.ID_TYPE: id_type,
            self.ID: identifier,
        })


class SystemAssignedManagedIdentity(ManagedIdentity):
    """Represent a system-assigned managed identity.

    It is equivalent to a Python dict of::

        {"ManagedIdentityIdType": "SystemAssigned", "Id": None}

    or a JSON blob of::

        {"ManagedIdentityIdType": "SystemAssigned", "Id": null}
    """
    def __init__(self):
        super(SystemAssignedManagedIdentity, self).__init__(id_type=self.SYSTEM_ASSIGNED)


class UserAssignedManagedIdentity(ManagedIdentity):
    """Represent a user-assigned managed identity.

    Depends on the id you provided, the outcome is equivalent to one of the below::

        {"ManagedIdentityIdType": "ClientId", "Id": "foo"}
        {"ManagedIdentityIdType": "ResourceId", "Id": "foo"}
        {"ManagedIdentityIdType": "ObjectId", "Id": "foo"}
    """
    def __init__(self, *, client_id=None, resource_id=None, object_id=None):
        if client_id and not resource_id and not object_id:
            super(UserAssignedManagedIdentity, self).__init__(
                id_type=self.CLIENT_ID, identifier=client_id)
        elif not client_id and resource_id and not object_id:
            super(UserAssignedManagedIdentity, self).__init__(
                id_type=self.RESOURCE_ID, identifier=resource_id)
        elif not client_id and not resource_id and object_id:
            super(UserAssignedManagedIdentity, self).__init__(
                id_type=self.OBJECT_ID, identifier=object_id)
        else:
            raise ManagedIdentityError(
                "You shall specify one of the three parameters: "
                "client_id, resource_id, object_id")


class _ThrottledHttpClient(ThrottledHttpClientBase):
    def __init__(self, *args, **kwargs):
        super(_ThrottledHttpClient, self).__init__(*args, **kwargs)
        self.get = IndividualCache(  # All MIs (except Cloud Shell) use GETs
            mapping=self._expiring_mapping,
            key_maker=lambda func, args, kwargs: "REQ {} hash={} 429/5xx/Retry-After".format(
                args[0],  # It is the endpoint, typically a constant per MI type
                self._hash(
                    # Managed Identity flavors have inconsistent parameters.
                    # We simply choose to hash them all.
                    str(kwargs.get("params")) + str(kwargs.get("data"))),
                ),
            expires_in=RetryAfterParser(5).parse,  # 5 seconds default for non-PCA
            )(self.get)  # Note: Decorate the parent get(), not the http_client.get()


class ManagedIdentityClient(object):
    """This API encapsulates multiple managed identity back-ends:
    VM, App Service, Azure Automation (Runbooks), Azure Function, Service Fabric,
    and Azure Arc.

    It also provides token cache support.

    .. note::

        Cloud Shell support is NOT implemented in this class.
        Since MSAL Python 1.18 in May 2022, it has been implemented in
        :func:`PublicClientApplication.acquire_token_interactive` via calling pattern
        ``PublicClientApplication(...).acquire_token_interactive(scopes=[...], prompt="none")``.
        That is appropriate, because Cloud Shell yields a token with
        delegated permissions for the end user who has signed in to the Azure Portal
        (like what a ``PublicClientApplication`` does),
        not a token with application permissions for an app.
    """
    __instance = "localhost"  # We used to get this value from socket.getfqdn()
        # but it is unreliable because getfqdn() either hangs or returns empty value
        # on some misconfigured machines
    _tenant = "managed_identity"
    _TOKEN_SOURCE = "token_source"
    _TOKEN_SOURCE_IDP = "identity_provider"
    _TOKEN_SOURCE_CACHE = "cache"

    def __init__(
        self,
        managed_identity: Union[
            dict,
            ManagedIdentity,  # Could use Type[ManagedIdentity] but it is deprecated in Python 3.9+
            SystemAssignedManagedIdentity,
            UserAssignedManagedIdentity,
            ],
        *,
        http_client,
        token_cache=None,
        http_cache=None,
        client_capabilities: Optional[List[str]] = None,
    ):
        """Create a managed identity client.

        :param managed_identity:
            It accepts an instance of :class:`SystemAssignedManagedIdentity`
            or :class:`UserAssignedManagedIdentity`.
            They are equivalent to a dict with a certain shape,
            which may be loaded from a JSON configuration file or an env var.

        :param http_client:
            An http client object. For example, you can use ``requests.Session()``,
            optionally with exponential backoff behavior demonstrated in this recipe::

                import msal, requests
                from requests.adapters import HTTPAdapter, Retry
                s = requests.Session()
                retries = Retry(total=3, backoff_factor=0.1, status_forcelist=[
                    429, 500, 501, 502, 503, 504])
                s.mount('https://', HTTPAdapter(max_retries=retries))
                managed_identity = ...
                client = msal.ManagedIdentityClient(managed_identity, http_client=s)

            For Service Fabric managed identity, ``http_client`` must be a
            ``requests.Session`` using the standard ``requests.adapters.HTTPAdapter``.
            MSAL derives a separate session for the Service Fabric endpoint so that
            its certificate thumbprint can be validated before the Secret header is sent.

        :param token_cache:
            Optional. It accepts a :class:`msal.TokenCache` instance to store tokens.
            It will use an in-memory token cache by default.

        :param http_cache:
            Optional. It has the same characteristics as the
            :paramref:`msal.ClientApplication.http_cache`.

        :param list[str] client_capabilities: (optional)
            Allows configuration of one or more client capabilities, e.g. ["CP1"].

            Client capability is meant to inform the Microsoft identity platform
            (STS) what this client is capable for,
            so STS can decide to turn on certain features.

            Implementation details:
            Client capability in Managed Identity is relayed as-is
            via ``xms_cc`` parameter on the wire.

        Recipe 1: Hard code a managed identity for your app::

            import msal, requests
            client = msal.ManagedIdentityClient(
                msal.UserAssignedManagedIdentity(client_id="foo"),
                http_client=requests.Session(),
                )
            token = client.acquire_token_for_client("resource")

        Recipe 2: Write once, run everywhere.
        If you use different managed identity on different deployment,
        you may use an environment variable (such as MY_MANAGED_IDENTITY_CONFIG)
        to store a json blob like
        ``{"ManagedIdentityIdType": "ClientId", "Id": "foo"}`` or
        ``{"ManagedIdentityIdType": "SystemAssigned", "Id": null}``.
        The following app can load managed identity configuration dynamically::

            import json, os, msal, requests
            config = os.getenv("MY_MANAGED_IDENTITY_CONFIG")
            assert config, "An ENV VAR with value should exist"
            client = msal.ManagedIdentityClient(
                json.loads(config),
                http_client=requests.Session(),
                )
            token = client.acquire_token_for_client("resource")
        """
        if not ManagedIdentity.is_managed_identity(managed_identity):
            raise ManagedIdentityError(
                f"Incorrect managed_identity: {managed_identity}")
        self._managed_identity = managed_identity
        self._http_client = _ThrottledHttpClient(
            # This class only throttles excess token acquisition requests.
            # It does not provide retry.
            # Retry is the http_client or caller's responsibility, not MSAL's.
            #
            # FWIW, here is the inconsistent retry recommendation.
            # 1. Only MI on VM defines exotic 404 and 410 retry recommendations
            #    ( https://learn.microsoft.com/en-us/entra/identity/managed-identities-azure-resources/how-to-use-vm-token#error-handling )
            #    (especially for 410 which was supposed to be a permanent failure).
            # 2. MI on Service Fabric specifically suggests to not retry on 404.
            #    ( https://learn.microsoft.com/en-us/azure/service-fabric/how-to-managed-cluster-managed-identity-service-fabric-app-code#error-handling )
            http_client,
            http_cache=http_cache,
        )
        self._token_cache = token_cache or TokenCache()
        self._client_capabilities = client_capabilities

    def acquire_token_for_client(
        self,
        *,
        resource: str,  # If/when we support scope, resource will become optional
        claims_challenge: Optional[str] = None,
    ):
        """Acquire token for the managed identity.

        The result will be automatically cached.
        Subsequent calls will automatically search from cache first.

        :param resource: The resource for which the token is acquired.

        :param claims_challenge:
            Optional.
            It is a string representation of a JSON object
            (which contains lists of claims being requested).

            The tenant admin may choose to revoke all Managed Identity tokens,
            and then a *claims challenge* will be returned by the target resource,
            as a `claims_challenge` directive in the `www-authenticate` header,
            even if the app developer did not opt in for the "CP1" client capability.
            Upon receiving a `claims_challenge`, MSAL will attempt to acquire a new token.

        .. note::

            Known issue: When an Azure VM has only one user-assigned managed identity,
            and your app specifies to use system-assigned managed identity,
            Azure VM may still return a token for your user-assigned identity.

            This is a service-side behavior that cannot be changed by this library.
            `Azure VM docs <https://learn.microsoft.com/en-us/azure/active-directory/managed-identities-azure-resources/how-to-use-vm-token#get-a-token-using-http>`_
        """
        access_token_to_refresh = None  # This could become a public parameter in the future
        access_token_from_cache = None
        client_id_in_cache = self._managed_identity.get(
            ManagedIdentity.ID, "SYSTEM_ASSIGNED_MANAGED_IDENTITY")
        now = time.time()
        if True:  # Attempt cache search even if receiving claims_challenge,
                  # because we want to locate the existing token (if any) and refresh it
            matches = self._token_cache.search(
                self._token_cache.CredentialType.ACCESS_TOKEN,
                target=[resource],
                query=dict(
                    client_id=client_id_in_cache,
                    environment=self.__instance,
                    realm=self._tenant,
                    home_account_id=None,
                ),
            )
            for entry in matches:
                expires_in = int(entry["expires_on"]) - now
                if expires_in < 5*60:  # Then consider it expired
                    continue  # Removal is not necessary, it will be overwritten
                if claims_challenge and not access_token_to_refresh:
                    # Since caller did not pinpoint the token causing claims challenge,
                    # we have to assume it is the first token we found in cache.
                    access_token_to_refresh = entry["secret"]
                    break
                logger.debug("Cache hit an AT")
                access_token_from_cache = {  # Mimic a real response
                    "access_token": entry["secret"],
                    "token_type": entry.get("token_type", "Bearer"),
                    "expires_in": int(expires_in),  # OAuth2 specs defines it as int
                    self._TOKEN_SOURCE: self._TOKEN_SOURCE_CACHE,
                }
                if "refresh_on" in entry:
                    access_token_from_cache["refresh_on"] = int(entry["refresh_on"])
                    if int(entry["refresh_on"]) < now:  # aging
                        break  # With a fallback in hand, we break here to go refresh
                return access_token_from_cache  # It is still good as new
        try:
            result = _obtain_token(
                self._http_client, self._managed_identity, resource,
                access_token_sha256_to_refresh=hashlib.sha256(
                    access_token_to_refresh.encode("utf-8")).hexdigest()
                    if access_token_to_refresh else None,
                client_capabilities=self._client_capabilities,
            )
            if "access_token" in result:
                expires_in = result.get("expires_in", 3600)
                if "refresh_in" not in result and expires_in >= 7200:
                    result["refresh_in"] = int(expires_in / 2)
                self._token_cache.add(dict(
                    client_id=client_id_in_cache,
                    scope=[resource],
                    token_endpoint="https://{}/{}".format(
                        self.__instance, self._tenant),
                    response=result,
                    params={},
                    data={},
                ))
                if "refresh_in" in result:
                    result["refresh_on"] = int(now + result["refresh_in"])
                result[self._TOKEN_SOURCE] = self._TOKEN_SOURCE_IDP
            if (result and "error" not in result) or (not access_token_from_cache):
                return result
        except:  # The exact HTTP exception is transportation-layer dependent
            # Typically network error. Potential AAD outage?
            if not access_token_from_cache:  # It means there is no fall back option
                raise  # We choose to bubble up the exception
        return access_token_from_cache


def _scope_to_resource(scope):  # This is an experimental reasonable-effort approach
    u = urlparse(scope)
    if u.scheme:
        return "{}://{}".format(u.scheme, u.netloc)
    return scope  # There is no much else we can do here


def _get_arc_endpoint():
    if "IDENTITY_ENDPOINT" in os.environ and "IMDS_ENDPOINT" in os.environ:
        return os.environ["IDENTITY_ENDPOINT"]
    if (  # Defined in https://eng.ms/docs/cloud-ai-platform/azure-core/azure-management-and-platforms/control-plane-bburns/hybrid-resource-provider/azure-arc-for-servers/specs/extension_authoring
        sys.platform == "linux" and os.path.exists("/opt/azcmagent/bin/himds")
        or sys.platform == "win32" and os.path.exists(os.path.expandvars(
            # Avoid Windows-only "%EnvVar%" syntax so that tests can be run on Linux
            r"${ProgramFiles}\AzureConnectedMachineAgent\himds.exe"
        ))
    ):
        return "http://localhost:40342/metadata/identity/oauth2/token"


APP_SERVICE = object()
AZURE_ARC = object()
CLOUD_SHELL = object()  # In MSAL Python, token acquisition was done by
    # PublicClientApplication(...).acquire_token_interactive(..., prompt="none")
MACHINE_LEARNING = object()
SERVICE_FABRIC = object()
DEFAULT_TO_VM = object()  # Unknown environment; default to VM; you may want to probe
def get_managed_identity_source():
    """Detect the current environment and return the likely identity source.

    When this function returns ``CLOUD_SHELL``, you should use
    :func:`msal.PublicClientApplication.acquire_token_interactive` with ``prompt="none"``
    to obtain a token.
    """
    if ("IDENTITY_ENDPOINT" in os.environ and "IDENTITY_HEADER" in os.environ
            and "IDENTITY_SERVER_THUMBPRINT" in os.environ
    ):
        return SERVICE_FABRIC
    if "IDENTITY_ENDPOINT" in os.environ and "IDENTITY_HEADER" in os.environ:
        return APP_SERVICE
    if "MSI_ENDPOINT" in os.environ and "MSI_SECRET" in os.environ:
        return MACHINE_LEARNING
    if _get_arc_endpoint():
        return AZURE_ARC
    if _is_running_in_cloud_shell():
        return CLOUD_SHELL
    return DEFAULT_TO_VM


def _obtain_token(
    http_client, managed_identity, resource,
    *,
    access_token_sha256_to_refresh: Optional[str] = None,
    client_capabilities: Optional[List[str]] = None,
):
    if ("IDENTITY_ENDPOINT" in os.environ and "IDENTITY_HEADER" in os.environ
            and "IDENTITY_SERVER_THUMBPRINT" in os.environ
    ):
        if managed_identity:
            logger.debug(
                "Ignoring managed_identity parameter. "
                "Managed Identity in Service Fabric is configured in the cluster, "
                "not during runtime. See also "
                "https://learn.microsoft.com/en-us/azure/service-fabric/configure-existing-cluster-enable-managed-identity-token-service")
        return _obtain_token_on_service_fabric(
            http_client,
            os.environ["IDENTITY_ENDPOINT"],
            os.environ["IDENTITY_HEADER"],
            os.environ["IDENTITY_SERVER_THUMBPRINT"],
            resource,
            access_token_sha256_to_refresh=access_token_sha256_to_refresh,
            client_capabilities=client_capabilities,
        )
    if "IDENTITY_ENDPOINT" in os.environ and "IDENTITY_HEADER" in os.environ:
        return _obtain_token_on_app_service(
            http_client,
            os.environ["IDENTITY_ENDPOINT"],
            os.environ["IDENTITY_HEADER"],
            managed_identity,
            resource,
        )
    if "MSI_ENDPOINT" in os.environ and "MSI_SECRET" in os.environ:
        # Back ported from https://github.com/Azure/azure-sdk-for-python/blob/azure-identity_1.15.0/sdk/identity/azure-identity/azure/identity/_credentials/azure_ml.py
        return _obtain_token_on_machine_learning(
            http_client,
            os.environ["MSI_ENDPOINT"],
            os.environ["MSI_SECRET"],
            managed_identity,
            resource,
        )
    arc_endpoint = _get_arc_endpoint()
    if arc_endpoint:
        return _obtain_token_on_arc(
            http_client, arc_endpoint, resource, managed_identity)
    return _obtain_token_on_azure_vm(http_client, managed_identity, resource)


def _adjust_param(params, managed_identity, types_mapping=None):
    # Modify the params dict in place
    id_name = (types_mapping or ManagedIdentity._types_mapping).get(
        managed_identity.get(ManagedIdentity.ID_TYPE))
    if id_name:
        params[id_name] = managed_identity[ManagedIdentity.ID]

def _obtain_token_on_azure_vm(http_client, managed_identity, resource):
    # Based on https://docs.microsoft.com/en-us/azure/active-directory/managed-identities-azure-resources/how-to-use-vm-token#get-a-token-using-http
    logger.debug("Obtaining token via managed identity on Azure VM")
    params = {
        "api-version": "2018-02-01",
        "resource": resource,
        }
    _adjust_param(params, managed_identity)
    resp = http_client.get(
        os.getenv(
            "AZURE_POD_IDENTITY_AUTHORITY_HOST", "http://169.254.169.254"
            ).strip("/") + "/metadata/identity/oauth2/token",
        params=params,
        headers={
            "Metadata": "true",
            "x-client-SKU": SKU,
            "x-client-Ver": __version__,
            "x-ms-client-request-id": str(uuid.uuid4()),
            },
        )
    try:
        payload = json.loads(resp.text)
        if payload.get("access_token") and payload.get("expires_in"):
            return {  # Normalizing the payload into OAuth2 format
                "access_token": payload["access_token"],
                "expires_in": int(payload["expires_in"]),
                "resource": payload.get("resource"),
                "token_type": payload.get("token_type", "Bearer"),
                }
        return payload  # It would be {"error": ..., "error_description": ...} according to https://learn.microsoft.com/en-us/entra/identity/managed-identities-azure-resources/how-to-use-vm-token#error-handling
    except json.decoder.JSONDecodeError:
        logger.debug("IMDS emits unexpected payload: %s", resp.text)
        raise

def _obtain_token_on_app_service(
    http_client, endpoint, identity_header, managed_identity, resource,
):
    """Obtains token for
    `App Service <https://learn.microsoft.com/en-us/azure/app-service/overview-managed-identity?tabs=portal%2Chttp#rest-endpoint-reference>`_,
    Azure Functions, and Azure Automation.
    """
    # Prerequisite: Create your app service https://docs.microsoft.com/en-us/azure/app-service/quickstart-python
    # Assign it a managed identity https://docs.microsoft.com/en-us/azure/app-service/overview-managed-identity?tabs=portal%2Chttp
    # SSH into your container for testing https://docs.microsoft.com/en-us/azure/app-service/configure-linux-open-ssh-session
    logger.debug("Obtaining token via managed identity on Azure App Service")
    params = {
        "api-version": "2019-08-01",
        "resource": resource,
        }
    _adjust_param(params, managed_identity, types_mapping={
        ManagedIdentity.CLIENT_ID: "client_id",
        ManagedIdentity.RESOURCE_ID: "mi_res_id",  # App Service's resource id uses "mi_res_id"
        ManagedIdentity.OBJECT_ID: "object_id",
    })

    resp = http_client.get(
        endpoint,
        params=params,
        headers={
            "X-IDENTITY-HEADER": identity_header,
            "Metadata": "true",  # Unnecessary yet harmless for App Service,
            # It will be needed by Azure Automation
            # https://docs.microsoft.com/en-us/azure/automation/enable-managed-identity-for-automation#get-access-token-for-system-assigned-managed-identity-using-http-get
            },
        )
    try:
        payload = json.loads(resp.text)
        if payload.get("access_token") and payload.get("expires_on"):
            return {  # Normalizing the payload into OAuth2 format
                "access_token": payload["access_token"],
                "expires_in": int(payload["expires_on"]) - int(time.time()),
                "resource": payload.get("resource"),
                "token_type": payload.get("token_type", "Bearer"),
                }
        return {
            "error": "invalid_scope",  # Empirically, wrong resource ends up with a vague statusCode=500
            "error_description": "{}, {}".format(
                payload.get("statusCode"), payload.get("message")),
            }
    except json.decoder.JSONDecodeError:
        logger.debug("IMDS emits unexpected payload: %s", resp.text)
        raise

def _obtain_token_on_machine_learning(
    http_client, endpoint, secret, managed_identity, resource,
):
    # Could not find protocol docs from https://docs.microsoft.com/en-us/azure/machine-learning
    # The following implementation is back ported from Azure Identity 1.15.0
    logger.debug("Obtaining token via managed identity on Azure Machine Learning")
    params = {"api-version": "2017-09-01", "resource": resource}
    _adjust_param(params, managed_identity)
    if params["api-version"] == "2017-09-01" and "client_id" in params:
        # Workaround for a known bug in Azure ML 2017 API
        params["clientid"] = params.pop("client_id")
    resp = http_client.get(
        endpoint,
        params=params,
        headers={"secret": secret},
        )
    try:
        payload = json.loads(resp.text)
        if payload.get("access_token") and payload.get("expires_on"):
            return {  # Normalizing the payload into OAuth2 format
                "access_token": payload["access_token"],
                "expires_in": int(payload["expires_on"]) - int(time.time()),
                "resource": payload.get("resource"),
                "token_type": payload.get("token_type", "Bearer"),
                }
        return {
            "error": "invalid_scope",  # TODO: To be tested
            "error_description": "{}".format(payload),
            }
    except json.decoder.JSONDecodeError:
        logger.debug("IMDS emits unexpected payload: %s", resp.text)
        raise


def _obtain_token_on_service_fabric(
    http_client, endpoint, identity_header, server_thumbprint, resource,
    *,
    access_token_sha256_to_refresh: str = None,
    client_capabilities: Optional[List[str]] = None,
):
    """Obtains token for
    `Service Fabric <https://learn.microsoft.com/en-us/azure/service-fabric/>`_
    """
    # Deployment https://learn.microsoft.com/en-us/azure/service-fabric/service-fabric-get-started-containers-linux
    # See also https://github.com/Azure/azure-sdk-for-python/blob/main/sdk/identity/azure-identity/tests/managed-identity-live/service-fabric/service_fabric.md
    # Protocol https://learn.microsoft.com/en-us/azure/service-fabric/how-to-managed-identity-service-fabric-app-code#acquiring-an-access-token-using-rest-api
    logger.debug("Obtaining token via managed identity on Azure Service Fabric")
    parsed_endpoint = urlparse(endpoint)
    if parsed_endpoint.scheme.lower() != "https" or not parsed_endpoint.hostname:
        raise ManagedIdentityError(
            "Service Fabric managed identity endpoint must use HTTPS.")
    service_fabric_http_client = _create_service_fabric_http_client(
        http_client, endpoint, _normalize_service_fabric_thumbprint(server_thumbprint))
    resp = service_fabric_http_client.get(
        endpoint,
        params={k: v for k, v in {
            "api-version": "2019-07-01-preview",
            "resource": resource,
            "token_sha256_to_refresh": access_token_sha256_to_refresh,
            "xms_cc": ",".join(client_capabilities) if client_capabilities else None,
            }.items() if v is not None},
        headers={"Secret": identity_header},
        )
    try:
        payload = json.loads(resp.text)
        if payload.get("access_token") and payload.get("expires_on"):
            return {  # Normalizing the payload into OAuth2 format
                "access_token": payload["access_token"],
                "expires_in": int(  # Despite the example in docs shows an integer,
                    payload["expires_on"]  # Azure SDK team's test obtained a string.
                    ) - int(time.time()),
                "resource": payload.get("resource"),
                "token_type": payload["token_type"],
                }
        error = payload.get("error", {})  # https://learn.microsoft.com/en-us/azure/service-fabric/how-to-managed-identity-service-fabric-app-code#error-handling
        error_mapping = {  # Map Service Fabric errors into OAuth2 errors  https://www.rfc-editor.org/rfc/rfc6749#section-5.2
            "SecretHeaderNotFound": "unauthorized_client",
            "ManagedIdentityNotFound": "invalid_client",
            "ArgumentNullOrEmpty": "invalid_scope",
            }
        return {
            "error": error_mapping.get(error.get("code"), "invalid_request"),
            "error_description": resp.text,
            }
    except json.decoder.JSONDecodeError:
        logger.debug("IMDS emits unexpected payload: %s", resp.text)
        raise


def _normalize_service_fabric_thumbprint(server_thumbprint):
    normalized = "".join(
        character for character in str(server_thumbprint)
        if character not in " \t\r\n:")
    if len(normalized) != 40 or any(
            character not in "0123456789abcdefABCDEF"
            for character in normalized):
        raise ManagedIdentityError(
            "IDENTITY_SERVER_THUMBPRINT must be a SHA-1 certificate thumbprint.")
    return normalized.lower()


class _ServiceFabricHTTPSConnection(HTTPSConnection):
    """An HTTPS connection that authenticates the Service Fabric endpoint certificate."""
    _server_thumbprint = None

    def connect(self):
        super(_ServiceFabricHTTPSConnection, self).connect()
        if getattr(self, "proxy_is_forwarding", False):
            self.close()
            raise ssl.SSLCertVerificationError(
                "Cannot validate the Service Fabric endpoint certificate through "
                "a forwarding proxy.")
        certificate = self.sock.getpeercert(binary_form=True)
        actual_thumbprint = hashlib.sha1(certificate).hexdigest()
        if not hmac.compare_digest(actual_thumbprint, self._server_thumbprint):
            self.close()
            raise ssl.SSLCertVerificationError(
                "Service Fabric endpoint certificate thumbprint does not match "
                "IDENTITY_SERVER_THUMBPRINT.")
        self.is_verified = True


class _ServiceFabricHTTPSConnectionPool(HTTPSConnectionPool):
    ConnectionCls = _ServiceFabricHTTPSConnection


class _ServiceFabricHTTPAdapter(HTTPAdapter):
    """Use certificate-thumbprint authentication for the Service Fabric endpoint."""

    def __init__(self, server_thumbprint, *args, **kwargs):
        connection_class = type(
            "_PinnedServiceFabricHTTPSConnection",
            (_ServiceFabricHTTPSConnection,),
            {"_server_thumbprint": server_thumbprint},
        )
        self._connection_pool_class = type(
            "_PinnedServiceFabricHTTPSConnectionPool",
            (_ServiceFabricHTTPSConnectionPool,),
            {"ConnectionCls": connection_class},
        )
        super(_ServiceFabricHTTPAdapter, self).__init__(*args, **kwargs)

    def _configure_pool_manager(self, pool_manager):
        # PoolManager's mapping is module-global by default, so copy it before
        # replacing HTTPS only for this derived Service Fabric session.
        pool_manager.pool_classes_by_scheme = pool_manager.pool_classes_by_scheme.copy()
        pool_manager.pool_classes_by_scheme["https"] = self._connection_pool_class

    def init_poolmanager(self, connections, maxsize, block=False, **pool_kwargs):
        super(_ServiceFabricHTTPAdapter, self).init_poolmanager(
            connections, maxsize, block=block, **pool_kwargs)
        self._configure_pool_manager(self.poolmanager)

    def proxy_manager_for(self, proxy, **proxy_kwargs):
        pool_manager = super(_ServiceFabricHTTPAdapter, self).proxy_manager_for(
            proxy, **proxy_kwargs)
        self._configure_pool_manager(pool_manager)
        return pool_manager

    def cert_verify(self, conn, url, verify, cert):
        # The exact Service Fabric certificate thumbprint is the trust anchor.
        # Do not inherit caller-provided verify=False or a custom CA configuration.
        super(_ServiceFabricHTTPAdapter, self).cert_verify(
            conn, url, verify=False, cert=cert)


def _create_service_fabric_http_client(http_client, endpoint, server_thumbprint):
    """Clone a standard Requests session and attach a pinning-only HTTPS transport.

    Custom HTTP clients and adapters are rejected because MSAL cannot prove that
    they will validate the certificate before transmitting the Secret header.
    """
    if isinstance(http_client, ThrottledHttpClientBase):
        http_client = http_client.http_client
    if not isinstance(http_client, requests.Session):
        raise ManagedIdentityError(
            "Service Fabric managed identity requires a requests.Session "
            "with the standard HTTPAdapter.")
    source_adapter = http_client.get_adapter(endpoint)
    if type(source_adapter) is not HTTPAdapter:
        raise ManagedIdentityError(
            "Service Fabric managed identity does not support custom HTTP adapters.")

    service_fabric_client = requests.Session()
    service_fabric_client.headers = http_client.headers.copy()
    service_fabric_client.cookies = http_client.cookies.copy()
    service_fabric_client.auth = http_client.auth
    service_fabric_client.params = copy.copy(http_client.params)
    service_fabric_client.hooks = {
        event: handlers[:] for event, handlers in http_client.hooks.items()}
    service_fabric_client.proxies = http_client.proxies.copy()
    service_fabric_client.stream = http_client.stream
    service_fabric_client.trust_env = http_client.trust_env
    service_fabric_client.max_redirects = http_client.max_redirects
    service_fabric_client.cert = http_client.cert
    service_fabric_client.verify = True
    service_fabric_client.adapters.clear()
    service_fabric_client.mount("https://", _ServiceFabricHTTPAdapter(
        server_thumbprint,
        max_retries=copy.deepcopy(source_adapter.max_retries),
        pool_connections=source_adapter._pool_connections,
        pool_maxsize=source_adapter._pool_maxsize,
        pool_block=source_adapter._pool_block,
    ))
    return service_fabric_client


_supported_arc_platforms_and_their_prefixes = {
    "linux": "/var/opt/azcmagent/tokens",
    "win32": os.path.expandvars(r"%ProgramData%\AzureConnectedMachineAgent\Tokens"),
}

class ArcPlatformNotSupportedError(ManagedIdentityError):
    pass

def _raise_if_arc_did_not_honor_user_assigned_identity(managed_identity, payload):
    """Fail closed when a user-assigned identity was requested but Azure Arc did not confirm it.

    A legacy Azure Arc agent ignores the client_id / object_id / msi_res_id selector and silently
    returns the machine's system-assigned identity. An agent that supports user-assigned managed
    identity echoes the identity it used in the token response. When that echo is missing or does
    not match the requested selector, MSAL must not hand back a token for a different identity than
    the one that was requested.
    """
    if not ManagedIdentity.is_user_assigned(managed_identity):
        return  # System-assigned: there is no requested identity to confirm
    requested = managed_identity.get(ManagedIdentity.ID)
    echoed = {
        ManagedIdentity.CLIENT_ID: payload.get("client_id"),
        ManagedIdentity.OBJECT_ID: payload.get("object_id"),
        # Azure Arc echoes msi_res_id; accept the mi_res_id spelling too as a safety net
        ManagedIdentity.RESOURCE_ID: payload.get("msi_res_id") or payload.get("mi_res_id"),
    }.get(managed_identity.get(ManagedIdentity.ID_TYPE))
    # Compare case-insensitively: client_id / object_id are GUIDs, and an ARM resource id
    # (msi_res_id) can legitimately differ in segment casing.
    if not echoed or str(echoed).lower() != str(requested).lower():
        raise ManagedIdentityError(
            "Azure Arc did not confirm the requested user-assigned managed identity "
            "in the token response. The agent likely does not support user-assigned "
            "managed identities and returned the system-assigned identity.")

def _obtain_token_on_arc(http_client, endpoint, resource, managed_identity=None):
    # https://learn.microsoft.com/en-us/azure/azure-arc/servers/managed-identity-authentication
    logger.debug("Obtaining token via managed identity on Azure Arc")
    params = {"api-version": "2020-06-01", "resource": resource}
    if managed_identity:
        _adjust_param(params, managed_identity, types_mapping={
            ManagedIdentity.CLIENT_ID: "client_id",
            ManagedIdentity.RESOURCE_ID: "msi_res_id",  # Azure Arc honors the IMDS msi_res_id spelling; mi_res_id is ignored and returns the system-assigned identity
            ManagedIdentity.OBJECT_ID: "object_id",
        })
    resp = http_client.get(
        endpoint,
        params=params.copy(),
        headers={"Metadata": "true"},
        )
    www_auth = "www-authenticate"  # Header in lower case
    challenge = {
        # Normalized to lowercase, because header names are case-insensitive
        # https://datatracker.ietf.org/doc/html/rfc7230#section-3.2
        k.lower(): v for k, v in resp.headers.items() if k.lower() == www_auth
        }.get(www_auth, "").split("=")  # Output will be ["Basic realm", "content"]
    if not (  # https://datatracker.ietf.org/doc/html/rfc7617#section-2
            len(challenge) == 2 and challenge[0].lower() == "basic realm"):
        raise ManagedIdentityError(
            "Unrecognizable WWW-Authenticate header: {}".format(resp.headers))
    if sys.platform not in _supported_arc_platforms_and_their_prefixes:
        raise ArcPlatformNotSupportedError(
            f"Platform {sys.platform} was undefined and unsupported")
    filename = os.path.join(
        # This algorithm is documented in an internal doc https://msazure.visualstudio.com/One/_wiki/wikis/One.wiki/233012/VM-Extension-Authoring-for-Arc?anchor=2.-obtaining-tokens
        _supported_arc_platforms_and_their_prefixes[sys.platform],
        os.path.splitext(os.path.basename(challenge[1]))[0] + ".key")
    if os.stat(filename).st_size > 4096:  # Check size BEFORE loading its content
        raise ManagedIdentityError("Local key file shall not be larger than 4KB")
    with open(filename) as f:
        secret = f.read()
    response = http_client.get(
        endpoint,
        params=params.copy(),
        headers={"Metadata": "true", "Authorization": "Basic {}".format(secret)},
        )
    try:
        payload = json.loads(response.text)
        if payload.get("access_token") and payload.get("expires_in"):
            # Example: https://learn.microsoft.com/en-us/azure/azure-arc/servers/media/managed-identity-authentication/bash-token-output-example.png
            _raise_if_arc_did_not_honor_user_assigned_identity(
                managed_identity, payload)
            return {
                "access_token": payload["access_token"],
                "expires_in": int(payload["expires_in"]),
                "token_type": payload.get("token_type", "Bearer"),
                "resource": payload.get("resource"),
                }
    except json.decoder.JSONDecodeError:
        pass
    return {
        "error": "invalid_request",
        "error_description": response.text,
        }



import os
import unittest
from types import SimpleNamespace
from unittest.mock import patch

from msal.region import (
    _detect_region, _detect_region_of_azure_vm, _validate_region)

from tests.http_client import MinimalResponse


class _StubHttpClient(object):
    """Records the requested URL/headers and returns a preconfigured response.

    If *response* is an exception instance, it is raised from ``get`` to
    simulate a network failure (e.g. not running in an Azure VM)."""

    def __init__(self, response):
        self._response = response
        self.url = None
        self.headers = None

    def get(self, url, params=None, headers=None, **kwargs):
        self.url = url
        self.headers = headers
        if isinstance(self._response, Exception):
            raise self._response
        return self._response



class TestValidateRegion(unittest.TestCase):

    def test_valid_simple_regions(self):
        for region in ("eastus", "westus2", "centralindia", "northeurope"):
            self.assertEqual(_validate_region(region), region)

    def test_valid_hyphenated_regions(self):
        for region in ("east-us", "west-us-2", "south-central-us"):
            self.assertEqual(_validate_region(region), region)

    def test_rejects_dots(self):
        self.assertIsNone(_validate_region("evil.com"))

    def test_rejects_slashes(self):
        self.assertIsNone(_validate_region("evil.com/path"))

    def test_rejects_colons(self):
        self.assertIsNone(_validate_region("evil:8080"))

    def test_rejects_uppercase(self):
        self.assertIsNone(_validate_region("EastUS"))

    def test_rejects_leading_hyphen(self):
        self.assertIsNone(_validate_region("-eastus"))

    def test_rejects_leading_digit(self):
        self.assertIsNone(_validate_region("2westus"))

    def test_rejects_trailing_hyphen(self):
        self.assertIsNone(_validate_region("westus-"))

    def test_rejects_label_longer_than_63_characters(self):
        self.assertIsNone(_validate_region("a" * 64))

    def test_rejects_adversarial_dns_labels(self):
        for region in (
                "EastUS", "east us", " eastus", "eastus ", "eastus\n",
                "westus2.login.microsoft.com", "https://westus2",
                "westus2:443", "westus2/path", "wéstus"):
            with self.subTest(region=region):
                self.assertIsNone(_validate_region(region))

    def test_empty_string(self):
        self.assertIsNone(_validate_region(""))

    def test_none(self):
        self.assertIsNone(_validate_region(None))


class TestDetectRegion(unittest.TestCase):

    @patch.dict(os.environ, {"REGION_NAME": "westus2"})
    def test_returns_valid_env_region(self):
        self.assertEqual(_detect_region(), "westus2")

    @patch.dict(os.environ, {"REGION_NAME": "evil.com/hijack"})
    def test_rejects_invalid_env_region(self):
        self.assertIsNone(_detect_region())

    def test_rejects_unnormalized_env_regions(self):
        for region in ("EastUS", "East Us 2", "westus-", "a" * 64):
            with self.subTest(region=region), patch.dict(
                    os.environ, {"REGION_NAME": region}):
                self.assertIsNone(_detect_region())

    @patch.dict(os.environ, {"REGION_NAME": ""})
    def test_empty_env_returns_none(self):
        self.assertIsNone(_detect_region())


class TestDetectRegionOfAzureVm(unittest.TestCase):

    def test_valid_location_is_returned(self):
        client = _StubHttpClient(
            MinimalResponse(status_code=200, text='{"location": "westus2"}'))
        self.assertEqual(_detect_region_of_azure_vm(client), "westus2")

    def test_request_uses_compute_json_endpoint(self):
        client = _StubHttpClient(
            MinimalResponse(status_code=200, text='{"location": "westus2"}'))
        _detect_region_of_azure_vm(client)
        self.assertEqual(
            client.url,
            "http://169.254.169.254/metadata/instance/compute"
            "?api-version=2021-02-01")
        self.assertNotIn("/location", client.url)
        self.assertNotIn("format=text", client.url)
        self.assertEqual(client.headers, {"Metadata": "true"})

    def test_missing_location_returns_none(self):
        client = _StubHttpClient(MinimalResponse(status_code=200, text="{}"))
        self.assertIsNone(_detect_region_of_azure_vm(client))

    def test_null_location_returns_none(self):
        client = _StubHttpClient(
            MinimalResponse(status_code=200, text='{"location": null}'))
        self.assertIsNone(_detect_region_of_azure_vm(client))

    def test_malformed_json_returns_none(self):
        client = _StubHttpClient(
            MinimalResponse(status_code=200, text="not json"))
        self.assertIsNone(_detect_region_of_azure_vm(client))

    def test_invalid_location_value_returns_none(self):
        client = _StubHttpClient(
            MinimalResponse(status_code=200, text='{"location": "evil.com/hijack"}'))
        self.assertIsNone(_detect_region_of_azure_vm(client))

    def test_invalid_dns_label_location_returns_none(self):
        for region in ("westus-", "a" * 64, "westus2:443", "wéstus"):
            with self.subTest(region=region):
                client = _StubHttpClient(MinimalResponse(
                    status_code=200, text='{"location": "%s"}' % region))
                self.assertIsNone(_detect_region_of_azure_vm(client))

    def test_non_string_location_returns_none(self):
        client = _StubHttpClient(
            MinimalResponse(status_code=200, text='{"location": 123}'))
        self.assertIsNone(_detect_region_of_azure_vm(client))

    def test_non_string_response_text_returns_none(self):
        # A custom http_client could yield a non-string resp.text; json.loads
        # would raise TypeError, which must be treated as a malformed response.
        client = _StubHttpClient(SimpleNamespace(status_code=200, text=None))
        self.assertIsNone(_detect_region_of_azure_vm(client))

    def test_network_failure_returns_none(self):
        client = _StubHttpClient(IOError("IMDS unreachable"))
        self.assertIsNone(_detect_region_of_azure_vm(client))


if __name__ == "__main__":
    unittest.main()



"""
This sample demonstrates a headless application that acquires a token using
the device code flow and then calls a web API with the token.
The configuration file would look like this:

This sample loads its configuration from a .env file.

To make this sample work, you need to choose one of the following templates:

    .env.sample.entra-id
    .env.sample.external-id
    .env.sample.external-id-with-custom-domain

Copy the chosen template to a new file named .env, and fill in the values.

You can then run this sample:

    python name_of_this_script.py
"""

import json
import logging
import os
import sys
import time

from dotenv import load_dotenv  # Need "pip install python-dotenv"
import msal
import requests


# Optional logging
# logging.basicConfig(level=logging.DEBUG)  # Enable DEBUG log for entire script
# logging.getLogger("msal").setLevel(logging.INFO)  # Optionally disable MSAL DEBUG logs

load_dotenv()  # We use this to load configuration from a .env file

# If for whatever reason you plan to recreate same ClientApplication periodically,
# you shall create one global token cache and reuse it by each ClientApplication
global_token_cache = msal.TokenCache()  # The TokenCache() is in-memory.
    # See more options in https://msal-python.readthedocs.io/en/latest/#tokencache

# Create a preferably long-lived app instance, to avoid the overhead of app creation
global_app = msal.PublicClientApplication(
    os.getenv('CLIENT_ID'),
    authority=os.getenv('AUTHORITY'),  # For Entra ID or External ID
    oidc_authority=os.getenv('OIDC_AUTHORITY'),  # For External ID with custom domain
    token_cache=global_token_cache,  # Let this app (re)use an existing token cache.
        # If absent, ClientApplication will create its own empty token cache
    )
scopes = os.getenv("SCOPE", "").split()


def acquire_and_use_token():
    # The pattern to acquire a token looks like this.
    result = None

    # Note: If your device-flow app does not have any interactive ability, you can
    #   completely skip the following cache part. But here we demonstrate it anyway.
    # We now check the cache to see if we have some end users signed in before.
    accounts = global_app.get_accounts()
    if accounts:
        logging.info("Account(s) exists in cache, probably with token too. Let's try.")
        print("Pick the account you want to use to proceed:")
        for a in accounts:
            print(a["username"])
        # Assuming the end user chose this one
        chosen = accounts[0]
        # Now let's try to find a token in cache for this account
        result = global_app.acquire_token_silent(scopes, account=chosen)

    if not result:
        logging.info("No suitable token exists in cache. Let's get a new one from AAD.")

        flow = global_app.initiate_device_flow(scopes=scopes)
        if "user_code" not in flow:
            raise ValueError(
                "Fail to create device flow. Err: %s" % json.dumps(flow, indent=4))

        print(flow["message"])
        sys.stdout.flush()  # Some terminal needs this to ensure the message is shown

        # Ideally you should wait here, in order to save some unnecessary polling
        # input("Press Enter after signing in from another device to proceed, CTRL+C to abort.")

        result = global_app.acquire_token_by_device_flow(flow)  # By default it will block
            # You can follow this instruction to shorten the block time
            #    https://msal-python.readthedocs.io/en/latest/#msal.PublicClientApplication.acquire_token_by_device_flow
            # or you may even turn off the blocking behavior,
            # and then keep calling acquire_token_by_device_flow(flow) in your own customized loop.

    if "access_token" in result:
        print("Token was obtained from:", result["token_source"])  # Since MSAL 1.25
        if os.getenv('ENDPOINT'):
            # Calling a web API using the access token
            api_result = requests.get(
                os.getenv('ENDPOINT'),
                headers={'Authorization': 'Bearer ' + result['access_token']},
                ).json()  # Assuming the response is JSON
            print("Web API call result", json.dumps(api_result, indent=2))
        else:
            print("Token acquisition result", json.dumps(result, indent=2))
    else:
        print("Token acquisition failed", result)  # Examine result["error_description"] etc. to diagnose error


while True:  # Here we mimic a long-lived daemon
    acquire_and_use_token()
    print("Press Ctrl-C to stop.")
    time.sleep(5)  # Let's say your app would run a workload every X minutes.
        # The first acquire_and_use_token() call will prompt. Others hit the cache.




"""End-to-end Managed Identity tests (real token acquisition).

These tests perform REAL token acquisition and therefore only run on the
self-hosted Azure DevOps pools that are actual Azure VM / Azure Arc machines with
the lab managed identities assigned:

  * IMDS tests -> the "MISEManagedIdentity" pool (an Azure VM). Gated on the
                  MSAL_TEST_MI_IMDS environment variable, which that pipeline
                  stage sets. (DEFAULT_TO_VM is also the fallback source on hosted
                  agents, so an explicit flag is used instead of source detection.)
  * Azure Arc  -> the "MISEAZUREARC" pool (an Azure Arc-enabled machine). Gated on
                  the Azure Arc source being detected on the machine.

They mirror the MSAL Go E2E tests
(apps/tests/e2e/managedidentity_e2e_test.go and managedidentity_arc_e2e_test.go)
and use the SAME lab identities and ARM resource, so both SDKs exercise the same
lab configuration on the same machines.

Everywhere else (hosted agents, local dev) the tests self-skip.
"""
import hashlib
import os
import unittest

import requests

from msal import (
    ManagedIdentityClient,
    SystemAssignedManagedIdentity,
    UserAssignedManagedIdentity,
)
from msal.managed_identity import get_managed_identity_source, AZURE_ARC


# Azure Resource Manager resource. Matches the ARM scope used by the MSAL .NET and
# Go managed identity E2E tests.
_ARM_RESOURCE = "https://management.azure.com"

# User-assigned managed identities assigned to the MISEManagedIdentity VM. These are
# the SAME values used by the MSAL Go / .NET IMDS E2E tests, so all SDKs exercise the
# same lab configuration on the same VM.
_UAMI_CLIENT_ID = "6325cd32-9911-41f3-819c-416cdf9104e7"
_UAMI_OBJECT_ID = "ecb2ad92-3e30-4505-b79f-ac640d069f24"
_UAMI_RESOURCE_ID = (
    "/subscriptions/c1686c51-b717-4fe0-9af3-24a20a41fb0c/resourcegroups/"
    "MSIV2-Testing-MSALNET/providers/Microsoft.ManagedIdentity/userAssignedIdentities/msiv2uami"
)


def _safe_error(result):
    """Return a log-safe summary of a failed result.

    Only the non-sensitive error fields are surfaced, so an assertion failure can
    never spill an access token (or the whole result dict) into the CI logs.
    """
    return {
        key: result[key]
        for key in ("error", "error_description", "correlation_id")
        if key in result
    }


def _acquire_token_twice_assert_caching(test, managed_identity):
    """Acquire an ARM token twice for the given managed identity and assert the first
    call reaches the identity provider while the second is served from the token cache.

    Shared by the IMDS and Azure Arc E2E tests, mirroring the Go helper of the same name.
    """
    http_client = requests.Session()
    client = ManagedIdentityClient(managed_identity, http_client=http_client)
    try:
        first = client.acquire_token_for_client(resource=_ARM_RESOURCE)
        test.assertNotIn(
            "error", first, "first acquisition failed: {}".format(_safe_error(first)))
        test.assertIn("access_token", first)
        test.assertEqual(
            "identity_provider", first.get("token_source"),
            "first call should reach the identity provider")

        second = client.acquire_token_for_client(resource=_ARM_RESOURCE)
        test.assertNotIn(
            "error", second, "second acquisition failed: {}".format(_safe_error(second)))
        test.assertIn("access_token", second)
        test.assertEqual(
            "cache", second.get("token_source"),
            "second call should be served from the token cache")
        # Compare tokens by SHA-256 digest so a mismatch never prints the actual
        # token material into CI logs.
        test.assertEqual(
            hashlib.sha256(first["access_token"].encode("utf-8")).hexdigest(),
            hashlib.sha256(second["access_token"].encode("utf-8")).hexdigest(),
            "cached token should match the original token")
    finally:
        http_client.close()


@unittest.skipUnless(
    os.getenv("MSAL_TEST_MI_IMDS"),
    "Set MSAL_TEST_MI_IMDS to run on the MISEManagedIdentity Azure VM (IMDS) pool")
class ManagedIdentityImdsE2ETestCase(unittest.TestCase):
    """Acquires ARM tokens over IMDS v1 for the system-assigned identity and each
    user-assigned identity binding (client id / resource id / object id). Each test
    asserts the first call reaches the identity provider and the second is cached."""

    def test_system_assigned(self):
        _acquire_token_twice_assert_caching(self, SystemAssignedManagedIdentity())

    def test_user_assigned_client_id(self):
        _acquire_token_twice_assert_caching(
            self, UserAssignedManagedIdentity(client_id=_UAMI_CLIENT_ID))

    def test_user_assigned_resource_id(self):
        _acquire_token_twice_assert_caching(
            self, UserAssignedManagedIdentity(resource_id=_UAMI_RESOURCE_ID))

    def test_user_assigned_object_id(self):
        _acquire_token_twice_assert_caching(
            self, UserAssignedManagedIdentity(object_id=_UAMI_OBJECT_ID))


@unittest.skipUnless(
    get_managed_identity_source() == AZURE_ARC,
    "Runs only on an Azure Arc-enabled machine (the MISEAZUREARC pool)")
class ManagedIdentityAzureArcE2ETestCase(unittest.TestCase):
    """Azure Arc supports the system-assigned identity only, so unlike the IMDS tests
    there are no user-assigned variants."""

    def test_system_assigned(self):
        _acquire_token_twice_assert_caching(self, SystemAssignedManagedIdentity())


if __name__ == "__main__":
    unittest.main()



import hashlib
import json
import os
import ssl
import sys
import tempfile
import threading
import time
import uuid
from datetime import datetime, timedelta, timezone
from http.server import BaseHTTPRequestHandler, ThreadingHTTPServer
from typing import List, Optional
import unittest
try:
    from unittest.mock import patch, ANY, mock_open, Mock
except:
    from mock import patch, ANY, mock_open, Mock
import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import SSLError
from urllib3.util.retry import Retry
from cryptography import x509
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID

from tests.test_throttled_http_client import (
    MinimalResponse, ThrottledHttpClientBaseTestCase, DummyHttpClient)
from msal import (
    SystemAssignedManagedIdentity, UserAssignedManagedIdentity,
    ManagedIdentityClient,
    ManagedIdentityError,
    ArcPlatformNotSupportedError,
)
from msal.managed_identity import (
    _ThrottledHttpClient,
    _supported_arc_platforms_and_their_prefixes,
    get_managed_identity_source,
    APP_SERVICE,
    AZURE_ARC,
    CLOUD_SHELL,
    MACHINE_LEARNING,
    SERVICE_FABRIC,
    DEFAULT_TO_VM,
    _create_service_fabric_http_client,
    _obtain_token_on_service_fabric,
)
from msal.token_cache import is_subdict_of

EXPECTED_SKU = "MSAL.Python"  # Hardcoded constant, not imported from product


class ManagedIdentityTestCase(unittest.TestCase):
    def test_helper_class_should_be_interchangable_with_dict_which_could_be_loaded_from_file_or_env_var(self):
        self.assertEqual(
            UserAssignedManagedIdentity(client_id="foo"),
            {"ManagedIdentityIdType": "ClientId", "Id": "foo"})
        self.assertEqual(
            UserAssignedManagedIdentity(resource_id="foo"),
            {"ManagedIdentityIdType": "ResourceId", "Id": "foo"})
        self.assertEqual(
            UserAssignedManagedIdentity(object_id="foo"),
            {"ManagedIdentityIdType": "ObjectId", "Id": "foo"})
        with self.assertRaises(ManagedIdentityError):
            UserAssignedManagedIdentity()
        with self.assertRaises(ManagedIdentityError):
            UserAssignedManagedIdentity(client_id="foo", resource_id="bar")
        self.assertEqual(
            SystemAssignedManagedIdentity(),
            {"ManagedIdentityIdType": "SystemAssigned", "Id": None})


class ThrottledHttpClientTestCase(ThrottledHttpClientBaseTestCase):
    def test_throttled_http_client_should_not_alter_original_http_client(self):
        self.assertNotAlteringOriginalHttpClient(_ThrottledHttpClient)

    def test_throttled_http_client_should_not_cache_successful_http_response(self):
        http_cache = {}
        http_client=DummyHttpClient(
            status_code=200,
            response_text='{"access_token": "AT", "expires_in": "1234", "resource": "R"}',
            )
        app = ManagedIdentityClient(
            SystemAssignedManagedIdentity(), http_client=http_client, http_cache=http_cache)
        result = app.acquire_token_for_client(resource="R")
        self.assertEqual("AT", result["access_token"])
        self.assertEqual({}, http_cache, "Should not cache successful http response")

    def test_throttled_http_client_should_cache_unsuccessful_http_response(self):
        http_cache = {}
        http_client=DummyHttpClient(
            status_code=400,
            response_headers={"Retry-After": "1"},
            response_text='{"error": "invalid_request"}',
            )
        app = ManagedIdentityClient(
            SystemAssignedManagedIdentity(), http_client=http_client, http_cache=http_cache)
        result = app.acquire_token_for_client(resource="R")
        self.assertEqual("invalid_request", result["error"])
        self.assertNotEqual({}, http_cache, "Should cache unsuccessful http response")
        self.assertCleanPickle(http_cache)


class ClientTestCase(unittest.TestCase):
    maxDiff = None

    def _build_app(
        self,
        *,
        client_capabilities: Optional[List[str]] = None,
    ):
        return ManagedIdentityClient(
            {   # Here we test it with the raw dict form, to test that
                # the client has no hard dependency on ManagedIdentity object
                "ManagedIdentityIdType": "SystemAssigned", "Id": None,
            },
            http_client=requests.Session(),
            client_capabilities=client_capabilities,
            )

    def setUp(self):
        self.app = self._build_app()

    def test_error_out_on_invalid_input(self):
        with self.assertRaises(ManagedIdentityError):
            ManagedIdentityClient({"foo": "bar"}, http_client=requests.Session())
        with self.assertRaises(ManagedIdentityError):
            ManagedIdentityClient(
                {"ManagedIdentityIdType": "undefined", "Id": "foo"},
                http_client=requests.Session())

    def assertCacheStatus(self, app):
        cache = app._token_cache._cache
        self.assertEqual(1, len(cache.get("AccessToken", [])), "Should have 1 AT")
        at = list(cache["AccessToken"].values())[0]
        self.assertEqual(
            app._managed_identity.get("Id", "SYSTEM_ASSIGNED_MANAGED_IDENTITY"),
            at["client_id"],
            "Should have expected client_id")
        self.assertEqual("managed_identity", at["realm"], "Should have expected realm")

    def _test_happy_path(
        self, app, mocked_http, expires_in, *, resource="R", claims_challenge=None,
    ):
        """It tests a normal token request that is expected to hit IdP,
        a subsequent same token request that is expected to hit cache,
        and then a request with claims_challenge that shall hit IdP again.
        """
        result = app.acquire_token_for_client(resource=resource)
        mocked_http.assert_called()
        call_count = mocked_http.call_count
        expected_result = {
            "access_token": "AT",
            "token_type": "Bearer",
        }
        self.assertTrue(
            is_subdict_of(expected_result, result),  # We will test refresh_on later
            "Should obtain a token response")
        self.assertTrue(result["token_source"], "identity_provider")
        self.assertEqual(expires_in, result["expires_in"], "Should have expected expires_in")
        if expires_in >= 7200:
            expected_refresh_on = int(time.time() + expires_in / 2)
            self.assertTrue(
                expected_refresh_on - 1 <= result["refresh_on"] <= expected_refresh_on + 1,
                "Should have a refresh_on time around the middle of the token's life")

        result = app.acquire_token_for_client(resource=resource)
        self.assertCacheStatus(app)
        self.assertEqual("cache", result["token_source"], "Should hit cache")
        self.assertEqual(
            call_count, mocked_http.call_count,
            "No new call to the mocked http should be made for a cache hit")
        self.assertTrue(
            is_subdict_of(expected_result, result),  # We will test refresh_on later
            "Should obtain a token response")
        self.assertTrue(
            expires_in - 5 < result["expires_in"] <= expires_in,
            "Should have similar expires_in")
        if expires_in >= 7200:
            self.assertTrue(
                expected_refresh_on - 5 < result["refresh_on"] <= expected_refresh_on,
                "Should have a refresh_on time around the middle of the token's life")

        result = app.acquire_token_for_client(
            resource=resource, claims_challenge=claims_challenge or "placeholder")
        self.assertEqual("identity_provider", result["token_source"], "Should miss cache")


class VmTestCase(ClientTestCase):

    def _test_happy_path(self) -> callable:
        expires_in = 7890  # We test a bigger than 7200 value here
        with patch.object(self.app._http_client, "get", return_value=MinimalResponse(
            status_code=200,
            text='{"access_token": "AT", "expires_in": "%s", "resource": "R"}' % expires_in,
        )) as mocked_method:
            super(VmTestCase, self)._test_happy_path(self.app, mocked_method, expires_in)
            return mocked_method

    def test_happy_path_of_vm(self):
        mock_get = self._test_happy_path()
        mock_get.assert_called_with(
            # The last call contained claims_challenge
            # but since IMDS doesn't support token_sha256_to_refresh,
            # the request shall remain the same as before
            'http://169.254.169.254/metadata/identity/oauth2/token',
            params={'api-version': '2018-02-01', 'resource': 'R'},
            headers={
                'Metadata': 'true',
                'x-client-SKU': EXPECTED_SKU,
                'x-client-Ver': ANY,
                'x-ms-client-request-id': ANY,
            },
            )
        # Validate correlation ID is a valid UUID
        corr_id = mock_get.call_args.kwargs["headers"]["x-ms-client-request-id"]
        uuid.UUID(corr_id)

    @patch.object(ManagedIdentityClient, "_ManagedIdentityClient__instance", "MixedCaseHostName")
    def test_happy_path_of_theoretical_mixed_case_hostname(self):
        """Historically, we used to get the host name from socket.getfqdn(),
        which could return a mixed-case host name on Windows.
        Although we no longer use getfqdn(), we still keep this test case to ensure we tolerate it.
        """
        self.test_happy_path_of_vm()

    @patch.dict(os.environ, {"AZURE_POD_IDENTITY_AUTHORITY_HOST": "http://localhost:1234//"})
    def test_happy_path_of_pod_identity(self):
        mock_get = self._test_happy_path()
        mock_get.assert_called_with(
            'http://localhost:1234/metadata/identity/oauth2/token',
            params={'api-version': '2018-02-01', 'resource': 'R'},
            headers={
                'Metadata': 'true',
                'x-client-SKU': EXPECTED_SKU,
                'x-client-Ver': ANY,
                'x-ms-client-request-id': ANY,
            },
            )
        # Validate correlation ID is a valid UUID
        corr_id = mock_get.call_args.kwargs["headers"]["x-ms-client-request-id"]
        uuid.UUID(corr_id)

    def test_vm_error_should_be_returned_as_is(self):
        raw_error = '{"raw": "error format is undefined"}'
        with patch.object(self.app._http_client, "get", return_value=MinimalResponse(
            status_code=400,
            text=raw_error,
        )) as mocked_method:
            self.assertEqual(
                json.loads(raw_error), self.app.acquire_token_for_client(resource="R"))
            self.assertEqual({}, self.app._token_cache._cache)

    def test_vm_resource_id_parameter_should_be_msi_res_id(self):
        app = ManagedIdentityClient(
            {"ManagedIdentityIdType": "ResourceId", "Id": "1234"},
            http_client=requests.Session(),
            )
        with patch.object(app._http_client, "get", return_value=MinimalResponse(
            status_code=200,
            text='{"access_token": "AT", "expires_in": 3600, "resource": "R"}',
        )) as mocked_method:
            app.acquire_token_for_client(resource="R")
            mocked_method.assert_called_with(
                'http://169.254.169.254/metadata/identity/oauth2/token',
                params={'api-version': '2018-02-01', 'resource': 'R', 'msi_res_id': '1234'},
                headers={
                    'Metadata': 'true',
                    'x-client-SKU': EXPECTED_SKU,
                    'x-client-Ver': ANY,
                    'x-ms-client-request-id': ANY,
                },
                )
            # Validate correlation ID is a valid UUID
            corr_id = mocked_method.call_args.kwargs["headers"]["x-ms-client-request-id"]
            uuid.UUID(corr_id)


@patch.dict(os.environ, {"IDENTITY_ENDPOINT": "http://localhost", "IDENTITY_HEADER": "foo"})
class AppServiceTestCase(ClientTestCase):

    def test_happy_path(self):
        expires_in = 1234
        with patch.object(self.app._http_client, "get", return_value=MinimalResponse(
            status_code=200,
            text='{"access_token": "AT", "expires_on": "%s", "resource": "R"}' % (
                int(time.time()) + expires_in),
        )) as mocked_method:
            self._test_happy_path(self.app, mocked_method, expires_in)

    def test_app_service_error_should_be_normalized(self):
        raw_error = '{"statusCode": 500, "message": "error content is undefined"}'
        with patch.object(self.app._http_client, "get", return_value=MinimalResponse(
            status_code=500,
            text=raw_error,
        )) as mocked_method:
            self.assertEqual({
                "error": "invalid_scope",
                "error_description": "500, error content is undefined",
            }, self.app.acquire_token_for_client(resource="R"))
            self.assertEqual({}, self.app._token_cache._cache)

    def test_app_service_resource_id_parameter_should_be_mi_res_id(self):
        app = ManagedIdentityClient(
            {"ManagedIdentityIdType": "ResourceId", "Id": "1234"},
            http_client=requests.Session(),
            )
        with patch.object(app._http_client, "get", return_value=MinimalResponse(
            status_code=200,
            text='{"access_token": "AT", "expires_on": 12345, "resource": "R"}',
        )) as mocked_method:
            app.acquire_token_for_client(resource="R")
            mocked_method.assert_called_with(
                'http://localhost',
                params={'api-version': '2019-08-01', 'resource': 'R', 'mi_res_id': '1234'},
                headers={'X-IDENTITY-HEADER': 'foo', 'Metadata': 'true'},
                )

    def test_past_expires_on_should_not_be_cached_long(self):
        # Regression test: a Managed Identity endpoint returning a PAST "expires_on"
        # timestamp must NOT result in a long-lived cached token. MSAL Python computes
        # expires_in = expires_on - now; a past timestamp yields a non-positive lifetime,
        # so any cached entry is immediately treated as expired and not served from cache.
        past_expires_on = int(time.time()) - 3600  # 1 hour in the past
        with patch.object(self.app._http_client, "get", return_value=MinimalResponse(
            status_code=200,
            text='{"access_token": "AT", "expires_on": "%s", "resource": "R"}' % past_expires_on,
        )) as mocked_method:
            first = self.app.acquire_token_for_client(resource="R")
            # Must not be inflated into a huge positive lifetime (a past absolute epoch
            # value must never be treated as a relative "seconds from now").
            self.assertLessEqual(
                first["expires_in"], 0,
                "A past expires_on must yield a non-positive expires_in, not an inflated lifetime")
            # A subsequent acquisition must re-hit the endpoint, proving the past-dated
            # token is treated as already expired and is not served from the cache.
            second = self.app.acquire_token_for_client(resource="R")
            self.assertEqual(
                "identity_provider", second["token_source"],
                "A token minted from a past expires_on must not be served from cache")
            self.assertEqual(
                2, mocked_method.call_count,
                "Second acquisition should re-call the MI endpoint, not reuse a stale cache entry")


@patch.dict(os.environ, {"MSI_ENDPOINT": "http://localhost", "MSI_SECRET": "foo"})
class MachineLearningTestCase(ClientTestCase):

    def test_happy_path(self):
        expires_in = 1234
        with patch.object(self.app._http_client, "get", return_value=MinimalResponse(
            status_code=200,
            text='{"access_token": "AT", "expires_on": "%s", "resource": "R"}' % (
                int(time.time()) + expires_in),
        )) as mocked_method:
            self._test_happy_path(self.app, mocked_method, expires_in)

    def test_machine_learning_error_should_be_normalized(self):
        raw_error = '{"error": "placeholder", "message": "placeholder"}'
        with patch.object(self.app._http_client, "get", return_value=MinimalResponse(
            status_code=500,
            text=raw_error,
        )) as mocked_method:
            self.assertEqual({
                "error": "invalid_scope",
                "error_description": "{'error': 'placeholder', 'message': 'placeholder'}",
            }, self.app.acquire_token_for_client(resource="R"))
            self.assertEqual({}, self.app._token_cache._cache)


@patch.dict(os.environ, {
    "IDENTITY_ENDPOINT": "https://localhost",
    "IDENTITY_HEADER": "foo",
    "IDENTITY_SERVER_THUMBPRINT": "ab" * 20,
})
class ServiceFabricTestCase(ClientTestCase):
    access_token = "AT"
    access_token_sha256 = hashlib.sha256(access_token.encode()).hexdigest()

    def _test_happy_path(self, app, *, claims_challenge=None) -> callable:
        expires_in = 1234
        with patch(
                "msal.managed_identity._create_service_fabric_http_client",
                return_value=app._http_client):
            with patch.object(app._http_client, "get", return_value=MinimalResponse(
                status_code=200,
                text='{"access_token": "%s", "expires_on": %s, "resource": "R", "token_type": "Bearer"}' % (
                    self.access_token, int(time.time()) + expires_in),
            )) as mocked_method:
                super(ServiceFabricTestCase, self)._test_happy_path(
                    app, mocked_method, expires_in, claims_challenge=claims_challenge)
                return mocked_method

    def test_happy_path_with_client_capabilities_should_relay_capabilities(self):
        self._test_happy_path(self._build_app(client_capabilities=["foo", "bar"])).assert_called_with(
            'https://localhost',
            params={
                'api-version': '2019-07-01-preview',
                'resource': 'R',
                'token_sha256_to_refresh': self.access_token_sha256,
                "xms_cc": "foo,bar",
            },
            headers={'Secret': 'foo'},
        )

    def test_happy_path_with_claim_challenge_should_send_sha256_to_provider(self):
        self._test_happy_path(
            self._build_app(client_capabilities=[]),  # Test empty client_capabilities
            claims_challenge='{"access_token": {"nbf": {"essential": true, "value": "1563308371"}}}',
        ).assert_called_with(
            'https://localhost',
            params={
                'api-version': '2019-07-01-preview',
                'resource': 'R',
                'token_sha256_to_refresh': self.access_token_sha256,
                # There is no xms_cc in this case
            },
            headers={'Secret': 'foo'},
        )

    def test_unified_api_service_should_ignore_unnecessary_client_id(self):
        self._test_happy_path(ManagedIdentityClient(
            {"ManagedIdentityIdType": "ClientId", "Id": "foo"},
            http_client=requests.Session(),
            ))

    def test_sf_error_should_be_normalized(self):
        raw_error = '''
{"error": {
    "correlationId": "foo",
    "code": "SecretHeaderNotFound",
    "message": "Secret is not found in the request headers."
}}'''  # https://learn.microsoft.com/en-us/azure/service-fabric/how-to-managed-identity-service-fabric-app-code#error-handling
        with patch(
                "msal.managed_identity._create_service_fabric_http_client",
                return_value=self.app._http_client):
            with patch.object(self.app._http_client, "get", return_value=MinimalResponse(
                status_code=404,
                text=raw_error,
            )) as mocked_method:
                self.assertEqual({
                    "error": "unauthorized_client",
                    "error_description": raw_error,
                }, self.app.acquire_token_for_client(resource="R"))
                self.assertEqual({}, self.app._token_cache._cache)


class _ServiceFabricTlsRequestHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        self.server.requests.append({
            "path": self.path,
            "headers": dict(self.headers),
        })
        body = json.dumps({
            "access_token": "AT",
            "expires_on": str(int(time.time()) + 3600),
            "resource": "R",
            "token_type": "Bearer",
        }).encode("utf-8")
        self.send_response(200)
        self.send_header("Content-Type", "application/json")
        self.send_header("Content-Length", str(len(body)))
        self.end_headers()
        self.wfile.write(body)

    def log_message(self, format, *args):
        pass


class ServiceFabricTlsValidationTestCase(unittest.TestCase):
    def setUp(self):
        self._temporary_directory = tempfile.TemporaryDirectory()
        private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)
        certificate = (
            x509.CertificateBuilder()
            .subject_name(x509.Name([
                x509.NameAttribute(NameOID.COMMON_NAME, "localhost"),
            ]))
            .issuer_name(x509.Name([
                x509.NameAttribute(NameOID.COMMON_NAME, "localhost"),
            ]))
            .public_key(private_key.public_key())
            .serial_number(x509.random_serial_number())
            .not_valid_before(datetime.now(timezone.utc))
            .not_valid_after(
                datetime.now(timezone.utc) + timedelta(days=1))
            .add_extension(
                x509.SubjectAlternativeName([x509.DNSName("localhost")]),
                critical=False,
            )
            .sign(private_key, hashes.SHA256())
        )
        self.thumbprint = certificate.fingerprint(hashes.SHA1()).hex()
        certificate_path = os.path.join(self._temporary_directory.name, "server.pem")
        private_key_path = os.path.join(self._temporary_directory.name, "server.key")
        with open(certificate_path, "wb") as certificate_file:
            certificate_file.write(certificate.public_bytes(serialization.Encoding.PEM))
        with open(private_key_path, "wb") as private_key_file:
            private_key_file.write(private_key.private_bytes(
                serialization.Encoding.PEM,
                serialization.PrivateFormat.TraditionalOpenSSL,
                serialization.NoEncryption(),
            ))
        self.server = ThreadingHTTPServer(
            ("localhost", 0), _ServiceFabricTlsRequestHandler)
        self.server.requests = []
        tls_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        tls_context.minimum_version = ssl.TLSVersion.TLSv1_2
        tls_context.load_cert_chain(certificate_path, private_key_path)
        self.server.socket = tls_context.wrap_socket(self.server.socket, server_side=True)
        self.server_thread = threading.Thread(target=self.server.serve_forever)
        self.server_thread.start()
        self.endpoint = "https://localhost:{}/token".format(self.server.server_port)

    def tearDown(self):
        self.server.shutdown()
        self.server.server_close()
        self.server_thread.join()
        self._temporary_directory.cleanup()

    def _new_session(self):
        session = requests.Session()
        session.trust_env = False
        return session

    def test_matching_thumbprint_sends_secret_after_validating_certificate(self):
        result = _obtain_token_on_service_fabric(
            self._new_session(),
            self.endpoint,
            "service-fabric-secret",
            ":".join(
                self.thumbprint[index:index + 2].upper()
                for index in range(0, len(self.thumbprint), 2)),
            "R",
        )

        self.assertEqual("AT", result["access_token"])
        self.assertEqual(1, len(self.server.requests))
        self.assertEqual(
            "service-fabric-secret",
            self.server.requests[0]["headers"]["Secret"])

    def test_mismatching_thumbprint_prevents_the_secret_from_being_sent(self):
        with self.assertRaises(SSLError):
            _obtain_token_on_service_fabric(
                self._new_session(),
                self.endpoint,
                "service-fabric-secret",
                "00" * 20,
                "R",
            )

        self.assertEqual([], self.server.requests)

    def test_non_https_endpoint_is_rejected_before_a_request_can_send_the_secret(self):
        with self.assertRaises(ManagedIdentityError):
            _obtain_token_on_service_fabric(
                self._new_session(),
                self.endpoint.replace("https://", "http://", 1),
                "service-fabric-secret",
                self.thumbprint,
                "R",
            )

        self.assertEqual([], self.server.requests)

    def test_derived_client_preserves_standard_session_settings_without_mutation(self):
        source = self._new_session()
        source.verify = False
        source.headers["X-Caller-Header"] = "caller-header"
        source.cookies.set("caller-cookie", "cookie-value")
        source.auth = ("caller", "password")
        source.params = {"caller-param": "caller-value"}
        source.proxies = {}
        source.max_redirects = 7
        source.mount("https://", HTTPAdapter(max_retries=Retry(total=2)))

        derived = _create_service_fabric_http_client(
            source, self.endpoint, self.thumbprint)

        self.assertFalse(source.verify)
        self.assertTrue(derived.verify)
        self.assertEqual(source.headers, derived.headers)
        self.assertEqual(source.cookies, derived.cookies)
        self.assertEqual(source.auth, derived.auth)
        self.assertEqual(source.params, derived.params)
        self.assertEqual(source.proxies, derived.proxies)
        self.assertEqual(source.max_redirects, derived.max_redirects)
        self.assertEqual(2, derived.get_adapter(self.endpoint).max_retries.total)
        self.assertIsNot(source.get_adapter(self.endpoint), derived.get_adapter(self.endpoint))
        response = derived.get(
            self.endpoint,
            params={"request-param": "request-value"},
            headers={"Secret": "service-fabric-secret"},
        )
        self.assertEqual(200, response.status_code)
        request = self.server.requests[0]
        self.assertIn("caller-param=caller-value", request["path"])
        self.assertIn("request-param=request-value", request["path"])
        self.assertEqual("caller-header", request["headers"]["X-Caller-Header"])
        self.assertIn("caller-cookie=cookie-value", request["headers"]["Cookie"])
        self.assertTrue(request["headers"]["Authorization"].startswith("Basic "))

    def test_custom_adapter_is_rejected_before_a_request_can_send_the_secret(self):
        source = self._new_session()

        class CustomAdapter(HTTPAdapter):
            pass

        source.mount("https://", CustomAdapter())
        with self.assertRaises(ManagedIdentityError):
            _obtain_token_on_service_fabric(
                source,
                self.endpoint,
                "service-fabric-secret",
                self.thumbprint,
                "R",
            )

        self.assertEqual([], self.server.requests)


@patch.dict(os.environ, {
    "IDENTITY_ENDPOINT": "http://localhost/token",
    "IMDS_ENDPOINT": "http://localhost",
})
@patch(
    "builtins.open" if sys.version_info.major >= 3 else "__builtin__.open",
    new=mock_open(read_data="secret"),  # `new` requires no extra argument on the decorated function.
        #  https://docs.python.org/3/library/unittest.mock.html#unittest.mock.patch
)
@patch("os.stat", return_value=Mock(st_size=4096))
class ArcTestCase(ClientTestCase):
    challenge = MinimalResponse(status_code=401, text="", headers={
        "WWW-Authenticate": "Basic realm=/tmp/foo",
        })

    def test_error_out_on_invalid_input(self, mocked_stat):
        return super(ArcTestCase, self).test_error_out_on_invalid_input()

    def test_happy_path(self, mocked_stat):
        expires_in = 1234
        with patch.object(self.app._http_client, "get", side_effect=[
            self.challenge,
            MinimalResponse(
                status_code=200,
                text='{"access_token": "AT", "expires_in": "%s", "resource": "R"}' % expires_in,
                ),
            ] * 2,  # Duplicate a pair of mocks for _test_happy_path()'s CAE check
        ) as mocked_method:
            try:
                self._test_happy_path(self.app, mocked_method, expires_in)
                mocked_stat.assert_called_with(os.path.join(
                    _supported_arc_platforms_and_their_prefixes[sys.platform],
                    "foo.key"))
            except ArcPlatformNotSupportedError:
                if sys.platform in _supported_arc_platforms_and_their_prefixes:
                    self.fail("Should not raise ArcPlatformNotSupportedError")

    def test_arc_error_should_be_normalized(self, mocked_stat):
        with patch.object(self.app._http_client, "get", side_effect=[
            self.challenge,
            MinimalResponse(status_code=400, text="undefined"),
        ]) as mocked_method:
            try:
                self.assertEqual({
                    "error": "invalid_request",
                    "error_description": "undefined",
                }, self.app.acquire_token_for_client(resource="R"))
                self.assertEqual({}, self.app._token_cache._cache)
            except ArcPlatformNotSupportedError:
                if sys.platform in _supported_arc_platforms_and_their_prefixes:
                    self.fail("Should not raise ArcPlatformNotSupportedError")

    def _assert_user_assigned_selector(self, managed_identity, selector_name, selector_value):
        app = ManagedIdentityClient(managed_identity, http_client=requests.Session())
        with patch.object(app._http_client, "get", side_effect=[
            self.challenge,
            MinimalResponse(
                status_code=200,
                # A compliant Azure Arc agent echoes the identity it used; MSAL verifies it (fail closed).
                text='{"access_token": "AT", "expires_in": "1234", "resource": "R", "%s": "%s"}' % (
                    selector_name, selector_value),
            ),
        ]) as mocked_method:
            try:
                result = app.acquire_token_for_client(resource="R")
                self.assertEqual("AT", result["access_token"])
                expected_params = {
                    "api-version": "2020-06-01",
                    "resource": "R",
                    selector_name: selector_value,
                }
                self.assertEqual(expected_params, mocked_method.call_args_list[0].kwargs["params"])
                self.assertEqual(expected_params, mocked_method.call_args_list[1].kwargs["params"])
            except ArcPlatformNotSupportedError:
                if sys.platform in _supported_arc_platforms_and_their_prefixes:
                    self.fail("Should not raise ArcPlatformNotSupportedError")

    def test_arc_user_assigned_client_id_should_be_forwarded(self, mocked_stat):
        self._assert_user_assigned_selector(
            UserAssignedManagedIdentity(client_id="client-id"),
            "client_id",
            "client-id",
        )

    def test_arc_user_assigned_resource_id_should_be_forwarded_as_msi_res_id(self, mocked_stat):
        self._assert_user_assigned_selector(
            UserAssignedManagedIdentity(resource_id="resource-id"),
            "msi_res_id",
            "resource-id",
        )

    def test_arc_user_assigned_object_id_should_be_forwarded(self, mocked_stat):
        self._assert_user_assigned_selector(
            UserAssignedManagedIdentity(object_id="object-id"),
            "object_id",
            "object-id",
        )

    def test_arc_user_assigned_identity_not_confirmed_should_fail_closed(self, mocked_stat):
        # A legacy Azure Arc agent ignores the selector and returns the system-assigned identity:
        # the token response echoes a different identity than the one requested. MSAL must fail
        # closed rather than hand back a token for a different identity than requested.
        app = ManagedIdentityClient(
            UserAssignedManagedIdentity(client_id="client-id"),
            http_client=requests.Session())
        with patch.object(app._http_client, "get", side_effect=[
            self.challenge,
            MinimalResponse(
                status_code=200,
                text='{"access_token": "AT", "expires_in": "1234", "resource": "R", "client_id": "a-different-id"}',
            ),
        ]):
            with self.assertRaises(ManagedIdentityError):
                app.acquire_token_for_client(resource="R")
        self.assertEqual(
            {}, app._token_cache._cache,
            "An unconfirmed identity token must not be cached")


class GetManagedIdentitySourceTestCase(unittest.TestCase):

    @patch.dict(os.environ, {
        "IDENTITY_ENDPOINT": "http://localhost",
        "IDENTITY_HEADER": "foo",
        "IDENTITY_SERVER_THUMBPRINT": "bar",
    })
    def test_service_fabric(self):
        self.assertEqual(get_managed_identity_source(), SERVICE_FABRIC)

    @patch.dict(os.environ, {
        "IDENTITY_ENDPOINT": "http://localhost",
        "IDENTITY_HEADER": "foo",
    })
    def test_app_service(self):
        self.assertEqual(get_managed_identity_source(), APP_SERVICE)

    @patch.dict(os.environ, {
        "MSI_ENDPOINT": "http://localhost",
        "MSI_SECRET": "foo",
    })
    def test_machine_learning(self):
        self.assertEqual(get_managed_identity_source(), MACHINE_LEARNING)

    @patch.dict(os.environ, {
        "IDENTITY_ENDPOINT": "http://localhost",
        "IMDS_ENDPOINT": "http://localhost",
    })
    def test_arc_by_env_var(self):
        self.assertEqual(get_managed_identity_source(), AZURE_ARC)

    @patch("msal.managed_identity.os.path.exists", return_value=True)
    @patch("msal.managed_identity.sys.platform", new="linux")
    def test_arc_by_file_existence_on_linux(self, mocked_exists):
        self.assertEqual(get_managed_identity_source(), AZURE_ARC)
        mocked_exists.assert_called_with("/opt/azcmagent/bin/himds")

    @patch("msal.managed_identity.os.path.exists", return_value=True)
    @patch("msal.managed_identity.sys.platform", new="win32")
    @patch.dict(os.environ, {"ProgramFiles": r"C:\Program Files"})
    def test_arc_by_file_existence_on_windows(self, mocked_exists):
        self.assertEqual(get_managed_identity_source(), AZURE_ARC)
        mocked_exists.assert_called_with(
            r"C:\Program Files\AzureConnectedMachineAgent\himds.exe")

    @patch.dict(os.environ, {
        "AZUREPS_HOST_ENVIRONMENT": "cloud-shell-foo",
    })
    def test_cloud_shell(self):
        self.assertEqual(get_managed_identity_source(), CLOUD_SHELL)

    def test_default_to_vm(self):
        self.assertEqual(get_managed_identity_source(), DEFAULT_TO_VM)



import json
import os
import logging
import re

logger = logging.getLogger(__name__)

_VALID_REGION_RE = re.compile(r"^[a-z](?:[a-z0-9-]{0,61}[a-z0-9])?$")

# IMDS compute metadata API version used for region auto-discovery.
# Bump this single constant when moving to a newer IMDS API version.
_IMDS_API_VERSION = "2021-02-01"


def _validate_region(region, source="unknown"):
    """Return *region* unchanged if it looks like a valid Azure region name,
    otherwise log a warning and return ``None``."""
    if region and _VALID_REGION_RE.fullmatch(region):
        return region
    if region:
        logger.warning(
            "Region from %s was discarded because it contains "
            "invalid characters: %r", source, region)
    return None


def _detect_region(http_client=None):
    region = os.environ.get("REGION_NAME", "")  # e.g. westus2
    if region:
        return _validate_region(region, source="REGION_NAME env variable")
    if http_client:
        return _detect_region_of_azure_vm(http_client)  # It could hang for minutes
    return None


def _detect_region_of_azure_vm(http_client):
    url = (
        "http://169.254.169.254/metadata/instance/compute"

        # The region is read from the "location" field of the compute metadata.
        # https://learn.microsoft.com/en-us/azure/virtual-machines/instance-metadata-service?tabs=linux#response-1
        "?api-version=" + _IMDS_API_VERSION
        )
    logger.info(
        "Connecting to IMDS {}. "
        "It may take a while if you are running outside of Azure. "
        "You should consider opting in/out region behavior on-demand, "
        'by loading a boolean flag "is_deployed_in_azure" '
        'from your per-deployment config and then do '
        '"app = ConfidentialClientApplication(..., '
        'azure_region=is_deployed_in_azure)"'.format(url))
    try:
        # https://docs.microsoft.com/en-us/azure/virtual-machines/windows/instance-metadata-service?tabs=linux#instance-metadata
        resp = http_client.get(url, headers={"Metadata": "true"})
    except:
        logger.info(
            "IMDS {} unavailable. Perhaps not running in Azure VM?".format(url))
        return None
    else:
        try:
            location = json.loads(resp.text).get("location")
        except (ValueError, AttributeError, TypeError):
            # ValueError: body is not valid JSON;
            # AttributeError: body is valid JSON but not a JSON object;
            # TypeError: resp.text is not a string (e.g. a custom http_client).
            logger.info("IMDS {} returned a malformed response.".format(url))
            return None
        if location is not None and not isinstance(location, str):
            logger.info("IMDS {} returned a non-string location.".format(url))
            return None
        return _validate_region(location, source="IMDS endpoint")



"""
This sample demonstrates a desktop application that acquires a token using a
pair of username and password, and then calls a web API with the token.

This sample loads its configuration from a .env file.

To make this sample work, you need to choose one of the following templates:

    .env.sample.entra-id
    .env.sample.external-id
    .env.sample.external-id-with-custom-domain

Copy the chosen template to a new file named .env, and fill in the values.

You can then run this sample:

    python name_of_this_script.py
"""

import getpass
import json
import logging
import os
import sys
import time

from dotenv import load_dotenv  # Need "pip install python-dotenv"
import msal
import requests


# Optional logging
# logging.basicConfig(level=logging.DEBUG)  # Enable DEBUG log for entire script
# logging.getLogger("msal").setLevel(logging.INFO)  # Optionally disable MSAL DEBUG logs

load_dotenv()  # We use this to load configuration from a .env file
if not os.getenv("USERNAME"):
    sys.exit("Please provide a username in the environment variable USERNAME.")
password = getpass.getpass("Password for {}: ".format(os.getenv("USERNAME")))

# If for whatever reason you plan to recreate same ClientApplication periodically,
# you shall create one global token cache and reuse it by each ClientApplication
global_token_cache = msal.TokenCache()  # The TokenCache() is in-memory.
    # See more options in https://msal-python.readthedocs.io/en/latest/#tokencache

# Create a preferably long-lived app instance, to avoid the overhead of app creation
global_app = msal.ClientApplication(
    os.getenv('CLIENT_ID'),
    authority=os.getenv('AUTHORITY'),  # For Entra ID or External ID
    oidc_authority=os.getenv('OIDC_AUTHORITY'),  # For External ID with custom domain
    client_credential=os.getenv('CLIENT_SECRET') or None,  # Treat empty string as None
    token_cache=global_token_cache,  # Let this app (re)use an existing token cache.
        # If absent, ClientApplication will create its own empty token cache
    )
scopes = os.getenv("SCOPE", "").split()


def acquire_and_use_token():
    # The pattern to acquire a token looks like this.
    result = None

    # Firstly, check the cache to see if this end user has signed in before
    accounts = global_app.get_accounts(username=os.getenv("USERNAME"))
    if accounts:
        logging.info("Account(s) exists in cache, probably with token too. Let's try.")
        result = global_app.acquire_token_silent(scopes, account=accounts[0])

    if not result:
        logging.info("No suitable token exists in cache. Let's get a new one from AAD.")
        # See this page for constraints of Username Password Flow.
        # https://github.com/AzureAD/microsoft-authentication-library-for-python/wiki/Username-Password-Authentication
        result = global_app.acquire_token_by_username_password(
            os.getenv("USERNAME"), password, scopes=scopes)

    if "access_token" in result:
        print("Token was obtained from:", result["token_source"])  # Since MSAL 1.25
        if os.getenv('ENDPOINT'):
            # Calling a web API using the access token
            api_result = requests.get(
                os.getenv('ENDPOINT'),
                headers={'Authorization': 'Bearer ' + result['access_token']},
                ).json()  # Assuming the response is JSON
            print("Web API call result", json.dumps(api_result, indent=2))
        else:
            print("Token acquisition result", json.dumps(result, indent=2))
    else:
        print("Token acquisition failed", result)  # Examine result["error_description"] etc. to diagnose error
        if 65001 in result.get("error_codes", []):  # Not mean to be coded programatically, but...
            raise RuntimeError(
                "Microsoft Entra ID requires user consent for U/P flow to succeed. "
                "Run acquire_token_interactive() instead.")


while True:  # Here we mimic a long-lived daemon
    acquire_and_use_token()
    print("Press Ctrl-C to stop.")
    time.sleep(5)  # Let's say your app would run a workload every X minutes.




import logging
import base64
import json
import time
import warnings

from msal.token_cache import (
    TokenCache, SerializableTokenCache, _compute_ext_cache_key,
    _parse_claims_or_raise, _merge_claims,
)
from tests import unittest


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


# NOTE: These helpers were once implemented as static methods in TokenCacheTestCase.
#       That would cause other test files' "from ... import TokenCacheTestCase"
#       to re-run all test cases in this file.
#       Now we avoid that, by defining these helpers in module level.
def build_id_token(
        iss="issuer", sub="subject", aud="my_client_id", exp=None, iat=None,
        **claims):  # AAD issues "preferred_username", ADFS issues "upn"
    return "header.%s.signature" % base64.b64encode(json.dumps(dict({
        "iss": iss,
        "sub": sub,
        "aud": aud,
        "exp": exp or (time.time() + 100),
        "iat": iat or time.time(),
        }, **claims)).encode()).decode('utf-8')


def build_response(  # simulate a response from AAD
        uid=None, utid=None,  # If present, they will form client_info
        access_token=None, expires_in=3600, token_type="some type",
        **kwargs  # Pass-through: refresh_token, foci, id_token, error, refresh_in, ...
        ):
    response = {}
    if uid and utid:  # Mimic the AAD behavior for "client_info=1" request
        response["client_info"] = base64.b64encode(json.dumps({
            "uid": uid, "utid": utid,
            }).encode()).decode('utf-8')
    if access_token:
        response.update({
            "access_token": access_token,
            "expires_in": expires_in,
            "token_type": token_type,
            })
    response.update(kwargs)  # Pass-through key-value pairs as top-level fields
    return response


class TokenCacheTestCase(unittest.TestCase):

    def setUp(self):
        self.cache = TokenCache()
        self.at_key_maker = self.cache.key_makers[
            TokenCache.CredentialType.ACCESS_TOKEN]

    def test_add_redacts_client_claims_in_debug_log(self):
        # Both fields live in event["data"] for different reasons: the
        # cache-key-only "client_claims" pseudo-param contributes to
        # ext_cache_key (it is excluded from the wire), while the merged "claims"
        # is a real OAuth wire parameter that is excluded from ext_cache_key. Both
        # may carry sensitive values, so TokenCache.add()'s DEBUG log must redact
        # them regardless of whether they affect the hash.
        secret = '{"access_token": {"nbf": {"essential": "SENSITIVE"}}}'
        with self.assertLogs("msal.token_cache", level="DEBUG") as cm:
            self.cache.add({
                "client_id": "my_client_id",
                "scope": ["s1"],
                "data": {"client_claims": secret, "claims": secret},
                "token_endpoint": "https://login.example.com/contoso/v2/token",
                "response": build_response(
                    uid="uid", utid="utid",
                    expires_in=3600, access_token="an access token"),
                }, now=1000)
        logged = "\n".join(cm.output)
        self.assertNotIn("SENSITIVE", logged)
        self.assertIn("********", logged)

    def testAddByAad(self):
        client_id = "my_client_id"
        id_token = build_id_token(
            oid="object1234", preferred_username="John Doe", aud=client_id)
        now = 1000
        self.cache.add({
            "client_id": client_id,
            "scope": ["s2", "s1", "s3"],  # Not in particular order
            "token_endpoint": "https://login.example.com/contoso/v2/token",
            "response": build_response(
                uid="uid", utid="utid",  # client_info
                expires_in=3600, access_token="an access token",
                id_token=id_token, refresh_token="a refresh token"),
            }, now=now)
        access_token_entry = {
                'cached_at': "1000",
                'client_id': 'my_client_id',
                'credential_type': 'AccessToken',
                'environment': 'login.example.com',
                'expires_on': "4600",
                'extended_expires_on': "4600",
                'home_account_id': "uid.utid",
                'realm': 'contoso',
                'secret': 'an access token',
                'target': 's1 s2 s3',  # Sorted
                'token_type': 'some type',
            }
        self.assertEqual(access_token_entry, self.cache._cache["AccessToken"].get(
            self.at_key_maker(**access_token_entry)))
        with warnings.catch_warnings():
            self.assertIn(
                access_token_entry,
                self.cache.find(self.cache.CredentialType.ACCESS_TOKEN, now=now),
                "find(..., query=None) should not crash, even though MSAL does not use it")
        self.assertEqual(
            {
                'client_id': 'my_client_id',
                'credential_type': 'RefreshToken',
                'environment': 'login.example.com',
                'home_account_id': "uid.utid",
                'last_modification_time': '1000',
                'secret': 'a refresh token',
                'target': 's1 s2 s3',  # Sorted
            },
            self.cache._cache["RefreshToken"].get(
                'uid.utid-login.example.com-refreshtoken-my_client_id--s1 s2 s3')
            )
        self.assertEqual(
            {
                'home_account_id': "uid.utid",
                'environment': 'login.example.com',
                'realm': 'contoso',
                'local_account_id': "object1234",
                'username': "John Doe",
                'authority_type': "MSSTS",
            },
            self.cache._cache["Account"].get('uid.utid-login.example.com-contoso')
            )
        self.assertEqual(
            {
                'credential_type': 'IdToken',
                'secret': id_token,
                'home_account_id': "uid.utid",
                'environment': 'login.example.com',
                'realm': 'contoso',
                'client_id': 'my_client_id',
            },
            self.cache._cache["IdToken"].get(
                'uid.utid-login.example.com-idtoken-my_client_id-contoso-')
            )
        self.assertEqual(
            {
                "client_id": "my_client_id",
                'environment': 'login.example.com',
            },
            self.cache._cache.get("AppMetadata", {}).get(
                "appmetadata-login.example.com-my_client_id")
            )

    def testAddByAdfs(self):
        client_id = "my_client_id"
        id_token = build_id_token(aud=client_id, upn="JaneDoe@example.com")
        self.cache.add({
            "client_id": client_id,
            "scope": ["s2", "s1", "s3"],  # Not in particular order
            "token_endpoint": "https://fs.msidlab8.com/adfs/oauth2/token",
            "response": build_response(
                uid=None, utid=None,  # ADFS will provide no client_info
                expires_in=3600, access_token="an access token",
                id_token=id_token, refresh_token="a refresh token"),
            }, now=1000)
        access_token_entry = {
                'cached_at': "1000",
                'client_id': 'my_client_id',
                'credential_type': 'AccessToken',
                'environment': 'fs.msidlab8.com',
                'expires_on': "4600",
                'extended_expires_on': "4600",
                'home_account_id': "subject",
                'realm': 'adfs',
                'secret': 'an access token',
                'target': 's1 s2 s3',  # Sorted
                'token_type': 'some type',
            }
        self.assertEqual(access_token_entry, self.cache._cache["AccessToken"].get(
            self.at_key_maker(**access_token_entry)))
        self.assertEqual(
            {
                'client_id': 'my_client_id',
                'credential_type': 'RefreshToken',
                'environment': 'fs.msidlab8.com',
                'home_account_id': "subject",
                'last_modification_time': "1000",
                'secret': 'a refresh token',
                'target': 's1 s2 s3',  # Sorted
            },
            self.cache._cache["RefreshToken"].get(
                'subject-fs.msidlab8.com-refreshtoken-my_client_id--s1 s2 s3')
            )
        self.assertEqual(
            {
                'home_account_id': "subject",
                'environment': 'fs.msidlab8.com',
                'realm': 'adfs',
                'local_account_id': "subject",
                'username': "JaneDoe@example.com",
                'authority_type': "ADFS",
            },
            self.cache._cache["Account"].get('subject-fs.msidlab8.com-adfs')
            )
        self.assertEqual(
            {
                'credential_type': 'IdToken',
                'secret': id_token,
                'home_account_id': "subject",
                'environment': 'fs.msidlab8.com',
                'realm': 'adfs',
                'client_id': 'my_client_id',
            },
            self.cache._cache["IdToken"].get(
                'subject-fs.msidlab8.com-idtoken-my_client_id-adfs-')
            )
        self.assertEqual(
            {
                "client_id": "my_client_id",
                'environment': 'fs.msidlab8.com',
            },
            self.cache._cache.get("AppMetadata", {}).get(
                "appmetadata-fs.msidlab8.com-my_client_id")
            )

    def assertFoundAccessToken(self, *, scopes, query, data=None, now=None):
        cached_at = None
        for cached_at in self.cache.search(
                TokenCache.CredentialType.ACCESS_TOKEN,
                target=scopes, query=query, now=now,
        ):
            for k, v in (data or {}).items():  # The extra data, if any
                self.assertEqual(cached_at.get(k), v, f"AT should contain {k}={v}")
        self.assertTrue(cached_at, "AT should be cached and searchable")
        return cached_at

    def _test_data_should_be_saved_and_searchable_in_access_token(self, data):
        scopes = ["s2", "s1", "s3"]  # Not in particular order
        now = 1000
        self.cache.add({
            "data": data,
            "client_id": "my_client_id",
            "scope": scopes,
            "token_endpoint": "https://login.example.com/contoso/v2/token",
            "response": build_response(
                uid="uid", utid="utid",  # client_info
                expires_in=3600, access_token="an access token",
                refresh_token="a refresh token"),
            }, now=now)
        self.assertFoundAccessToken(scopes=scopes, data=data, now=now, query=dict(
            data,  # Also use the extra data as a query criteria
            client_id="my_client_id",
            environment="login.example.com",
            realm="contoso",
            home_account_id="uid.utid",
        ))

    def test_extra_data_should_also_be_recorded_and_searchable_in_access_token(self):
        self._test_data_should_be_saved_and_searchable_in_access_token({"key_id": "1"})

    def test_access_tokens_with_different_key_id(self):
        self._test_data_should_be_saved_and_searchable_in_access_token({"key_id": "1"})
        self._test_data_should_be_saved_and_searchable_in_access_token({"key_id": "2"})
        self.assertEqual(
            len(self.cache._cache["AccessToken"]),
            1, """Historically, tokens are not keyed by key_id,
so a new token overwrites the old one, and we would end up with 1 token in cache""")

    def test_refresh_in_should_be_recorded_as_refresh_on(self):  # Sounds weird. Yep.
        scopes = ["s2", "s1", "s3"]  # Not in particular order
        self.cache.add({
            "client_id": "my_client_id",
            "scope": scopes,
            "token_endpoint": "https://login.example.com/contoso/v2/token",
            "response": build_response(
                uid="uid", utid="utid",  # client_info
                expires_in=3600, refresh_in=1800, access_token="an access token",
                ),  #refresh_token="a refresh token"),
            }, now=1000)
        at = self.assertFoundAccessToken(scopes=scopes, query=dict(
            client_id="my_client_id",
            environment="login.example.com",
            realm="contoso",
            home_account_id="uid.utid",
        ))
        self.assertEqual("2800", at.get("refresh_on"), "Should save refresh_on")

    def test_old_rt_data_with_wrong_key_should_still_be_salvaged_into_new_rt(self):
        sample = {
            'client_id': 'my_client_id',
            'credential_type': 'RefreshToken',
            'environment': 'login.example.com',
            'home_account_id': "uid.utid",
            'secret': 'a refresh token',
            'target': 's2 s1 s3',
            }
        new_rt = "this is a new RT"
        self.cache._cache["RefreshToken"] = {"wrong-key": sample}
        self.cache.modify(
            self.cache.CredentialType.REFRESH_TOKEN, sample, {"secret": new_rt})
        self.assertEqual(
            dict(sample, secret=new_rt),
            self.cache._cache["RefreshToken"].get(
                'uid.utid-login.example.com-refreshtoken-my_client_id--s2 s1 s3')
            )


class SerializableTokenCacheTestCase(unittest.TestCase):
    # Run all inherited test methods, and have extra check in tearDown()

    def setUp(self):
        self.cache = SerializableTokenCache()
        self.cache.deserialize("""
            {
                "AccessToken": {
                    "an-entry": {
                        "foo": "bar"
                        }
                    },
                "customized": "whatever"
            }
            """)

    def test_has_state_changed(self):
        cache = SerializableTokenCache()
        self.assertFalse(cache.has_state_changed)
        cache.add({})  # An NO-OP add() still counts as a state change. Good enough.
        self.assertTrue(cache.has_state_changed)

    def tearDown(self):
        state = self.cache.serialize()
        logger.debug("serialize() = %s", state)
        # Now assert all extended content are kept intact
        output = json.loads(state)
        self.assertEqual(output.get("customized"), "whatever",
            "Undefined cache keys and their values should be intact")
        self.assertEqual(
            output.get("AccessToken", {}).get("an-entry"), {"foo": "bar"},
            "Undefined token keys and their values should be intact")


class TestComputeExtCacheKey(unittest.TestCase):
    """Tests for the _compute_ext_cache_key hash function."""

    def test_empty_data_returns_empty_string(self):
        self.assertEqual("", _compute_ext_cache_key(None))
        self.assertEqual("", _compute_ext_cache_key({}))

    def test_excluded_fields_are_ignored(self):
        self.assertEqual("", _compute_ext_cache_key({"key_id": "k1", "token_type": "ssh-cert", "req_cnf": "nonce", "claims": "{}"}),
            "Fields in _EXT_CACHE_KEY_EXCLUDED_FIELDS should produce an empty hash")

    def test_fmi_path_produces_non_empty_hash(self):
        result = _compute_ext_cache_key({"fmi_path": "SomePath/Credential"})
        self.assertNotEqual("", result)
        self.assertIsInstance(result, str)

    def test_same_input_produces_same_hash(self):
        h1 = _compute_ext_cache_key({"fmi_path": "path/a"})
        h2 = _compute_ext_cache_key({"fmi_path": "path/a"})
        self.assertEqual(h1, h2)

    def test_different_fmi_paths_produce_different_hashes(self):
        h1 = _compute_ext_cache_key({"fmi_path": "path/a"})
        h2 = _compute_ext_cache_key({"fmi_path": "path/b"})
        self.assertNotEqual(h1, h2)

    def test_empty_fmi_path_value_is_ignored(self):
        self.assertEqual("", _compute_ext_cache_key({"fmi_path": ""}))

    def test_excluded_fields_dont_affect_hash(self):
        h1 = _compute_ext_cache_key({"fmi_path": "path/a"})
        h2 = _compute_ext_cache_key({"fmi_path": "path/a", "key_id": "k1", "req_cnf": "nonce"})
        self.assertEqual(h1, h2, "Excluded fields should not affect the hash")

    def test_non_excluded_fields_are_included_in_hash(self):
        h1 = _compute_ext_cache_key({"fmi_path": "path/a"})
        h2 = _compute_ext_cache_key({"fmi_path": "path/a", "custom_param": "val"})
        self.assertNotEqual(h1, h2, "Non-excluded fields should change the hash")


class TestClientClaimsCacheKey(unittest.TestCase):
    """client_claims must drive the extended cache key; server-issued claims must not."""

    def test_client_claims_produces_non_empty_hash(self):
        result = _compute_ext_cache_key({"client_claims": '{"a": 1}'})
        self.assertNotEqual("", result)
        self.assertIsInstance(result, str)

    def test_claims_is_excluded_but_client_claims_is_not(self):
        # Server-issued "claims" (claims_challenge) must not affect the key, because
        # it bypasses the cache. Client-originated "client_claims" must affect it.
        self.assertEqual("", _compute_ext_cache_key({"claims": '{"a": 1}'}))
        self.assertNotEqual("", _compute_ext_cache_key({"client_claims": '{"a": 1}'}))

    def test_same_client_claims_produce_same_hash(self):
        self.assertEqual(
            _compute_ext_cache_key({"client_claims": '{"a": 1}'}),
            _compute_ext_cache_key({"client_claims": '{"a": 1}'}))

    def test_different_client_claims_produce_different_hashes(self):
        self.assertNotEqual(
            _compute_ext_cache_key({"client_claims": '{"a": 1}'}),
            _compute_ext_cache_key({"client_claims": '{"a": 2}'}))

    def test_empty_client_claims_value_is_ignored(self):
        self.assertEqual("", _compute_ext_cache_key({"client_claims": ""}))


class TestClaimsHelpers(unittest.TestCase):
    """Tests for the shared _parse_claims_or_raise / _merge_claims helpers."""

    def test_parse_valid_object(self):
        self.assertEqual({"a": 1}, _parse_claims_or_raise('{"a": 1}'))

    def test_parse_rejects_non_object_and_malformed(self):
        for bad in ["not json", "[1, 2]", "null", "123", '"a string"', "true"]:
            with self.assertRaises(ValueError, msg="{!r} should raise".format(bad)):
                _parse_claims_or_raise(bad)

    def test_parse_rejects_non_string_types(self):
        # Non-str/bytes inputs make json.loads raise TypeError; the helper must
        # surface the same friendly ValueError so callers behave consistently
        # regardless of the bad input's type.
        for bad in [123, None, 1.5, True, ["a"], {"a": 1}]:
            with self.assertRaises(
                    ValueError, msg="{!r} should raise ValueError".format(bad)):
                _parse_claims_or_raise(bad)

    def test_parse_error_does_not_leak_raw_claims(self):
        # A malformed payload that contains a secret-looking value
        secret = '{"super": "secret-value-123"'  # missing closing brace
        with self.assertRaises(ValueError) as ctx:
            _parse_claims_or_raise(secret)
        self.assertNotIn("secret-value-123", str(ctx.exception),
            "Error message must never echo the raw claims content")

    def test_merge_returns_other_when_one_side_is_empty(self):
        self.assertEqual({"a": 1}, json.loads(_merge_claims(None, '{"a": 1}')))
        self.assertEqual({"a": 1}, json.loads(_merge_claims('{"a": 1}', None)))
        self.assertEqual({"a": 1}, json.loads(_merge_claims("", '{"a": 1}')))
        self.assertEqual({"a": 1}, json.loads(_merge_claims('{"a": 1}', "")))

    def test_merge_of_two_empties_is_falsy(self):
        self.assertFalse(_merge_claims(None, None))
        self.assertFalse(_merge_claims("", ""))

    def test_merge_deep_merges_objects(self):
        merged = json.loads(_merge_claims(
            '{"access_token": {"xms_cc": {"values": ["cp1"]}}}',
            '{"access_token": {"xms_az_nwperimid": {"essential": true}}}'))
        self.assertEqual(
            {"xms_cc": {"values": ["cp1"]}, "xms_az_nwperimid": {"essential": True}},
            merged["access_token"])

    def test_merge_leaf_conflict_second_arg_wins(self):
        # When both sides set the SAME leaf to different values, the second
        # argument (client-originated claims) wins. _acquire_token_for_client
        # passes server-issued claims first and forwarded_client_claims second,
        # so the caller's value takes precedence on a direct conflict.
        merged = json.loads(_merge_claims(
            '{"access_token": {"acrs": {"values": ["server"]}}}',
            '{"access_token": {"acrs": {"values": ["client"]}}}'))
        self.assertEqual({"values": ["client"]}, merged["access_token"]["acrs"])

    def test_merge_nested_conflict_preserves_disjoint_siblings(self):
        # A conflict on one nested key must not drop the other side's disjoint
        # sibling keys at the same level.
        merged = json.loads(_merge_claims(
            '{"access_token": {"a": {"values": ["server"]}, "keep_server": 1}}',
            '{"access_token": {"a": {"values": ["client"]}, "keep_client": 2}}'))
        self.assertEqual(
            {"a": {"values": ["client"]}, "keep_server": 1, "keep_client": 2},
            merged["access_token"])


class TestExtCacheKeyIsolation(unittest.TestCase):
    """Tests that ext_cache_key provides proper cache isolation in TokenCache."""

    def _build_event(self, client_id, scope, token_endpoint, access_token, data=None, **kwargs):
        return {
            "client_id": client_id,
            "scope": scope,
            "token_endpoint": token_endpoint,
            "response": build_response(access_token=access_token, expires_in=3600),
            "data": data or {},
            **kwargs,
        }

    def test_at_key_includes_ext_cache_key_when_present(self):
        cache = TokenCache()
        key_maker = cache.key_makers[TokenCache.CredentialType.ACCESS_TOKEN]
        key_without = key_maker(
            home_account_id="", environment="env", client_id="cid",
            realm="realm", target="scope")
        key_with = key_maker(
            home_account_id="", environment="env", client_id="cid",
            realm="realm", target="scope", ext_cache_key="somehash")
        self.assertNotEqual(key_without, key_with,
            "Keys with and without ext_cache_key should differ")
        self.assertIn("somehash", key_with)

    def test_different_ext_cache_keys_produce_different_at_keys(self):
        cache = TokenCache()
        key_maker = cache.key_makers[TokenCache.CredentialType.ACCESS_TOKEN]
        key_a = key_maker(
            home_account_id="", environment="env", client_id="cid",
            realm="realm", target="scope", ext_cache_key="hash_a")
        key_b = key_maker(
            home_account_id="", environment="env", client_id="cid",
            realm="realm", target="scope", ext_cache_key="hash_b")
        self.assertNotEqual(key_a, key_b)

    def test_fmi_tokens_are_stored_with_ext_cache_key(self):
        cache = TokenCache()
        event = self._build_event(
            "cid", ["s1"], "https://login.example.com/tenant/v2/token",
            "fmi_token", data={"fmi_path": "some/path"})
        cache.add(event)
        at_entries = list(cache.search(TokenCache.CredentialType.ACCESS_TOKEN, target=["s1"]))
        self.assertEqual(0, len(at_entries),
            "FMI tokens should NOT be found by a query without ext_cache_key")

    def test_fmi_tokens_found_with_matching_ext_cache_key_query(self):
        cache = TokenCache()
        ext_key = _compute_ext_cache_key({"fmi_path": "some/path"})
        event = self._build_event(
            "cid", ["s1"], "https://login.example.com/tenant/v2/token",
            "fmi_token", data={"fmi_path": "some/path"})
        cache.add(event)
        at_entries = list(cache.search(
            TokenCache.CredentialType.ACCESS_TOKEN, target=["s1"],
            query={"client_id": "cid", "environment": "login.example.com",
                   "realm": "tenant", "home_account_id": None,
                   "ext_cache_key": ext_key}))
        self.assertEqual(1, len(at_entries))
        self.assertEqual("fmi_token", at_entries[0]["secret"])

    def test_non_fmi_tokens_not_affected_by_fmi_cache(self):
        cache = TokenCache()
        # Add FMI token
        cache.add(self._build_event(
            "cid", ["s1"], "https://login.example.com/tenant/v2/token",
            "fmi_token", data={"fmi_path": "some/path"}))
        # Add regular token
        cache.add(self._build_event(
            "cid", ["s1"], "https://login.example.com/tenant/v2/token",
            "regular_token"))
        # Search without ext_cache_key should find only regular token
        at_entries = list(cache.search(
            TokenCache.CredentialType.ACCESS_TOKEN, target=["s1"],
            query={"client_id": "cid", "environment": "login.example.com",
                   "realm": "tenant", "home_account_id": None}))
        self.assertEqual(1, len(at_entries))
        self.assertEqual("regular_token", at_entries[0]["secret"])


class TestCrossMsalCacheKeyCompatibility(unittest.TestCase):
    """Verify that _compute_ext_cache_key produces hashes identical to the shared
    MSAL length-prefix ("netstring") encoding used by MSAL Go's
    CacheExtKeyGenerator (AzureAD/microsoft-authentication-library-for-go#629).

    The algorithm:
      1. Sort key-value pairs alphabetically by key (ordinal / case-sensitive)
      2. Length-prefix each part with its UTF-8 byte length and concatenate:
         "<len(key)>:<key><len(value)>:<value>..."
      3. SHA-256 hash
      4. Base64url encode (no padding), lowercased

    The length prefixes make the *serialization* injective: distinct component
    sets can never produce the same pre-hash string. (The final cache key is a
    SHA-256 digest, so only a cryptographically negligible hash collision
    remains possible.)

    NOTE: This is the encoding the whole MSAL SDK family is converging on
    (Go already merged it; .NET/Java/JS are landing the same change), so these
    hashes are byte-identical across SDKs. The cache *key format* (the 'atext'
    segment layout, asserted below) matches Go and .NET too. Caches are not
    shared across languages, so cross-language parity is a convenience, not a
    correctness requirement.
    """

    def test_two_params_hash_matches_shared_encoding(self):
        """Shared length-prefix expected: latlwkpewb_a0rcsmjvkecqt0_huumkw4sflzociike"""
        result = _compute_ext_cache_key({"key1": "value1", "key2": "value2"})
        self.assertEqual("latlwkpewb_a0rcsmjvkecqt0_huumkw4sflzociike", result)

    def test_two_different_params_hash_matches_shared_encoding(self):
        """Shared length-prefix expected: jjoe9jgfmdtnj0rzuetsqy7kzs2m1xfnjjxwsfxsrxq"""
        result = _compute_ext_cache_key({"key3": "value3", "key4": "value4"})
        self.assertEqual("jjoe9jgfmdtnj0rzuetsqy7kzs2m1xfnjjxwsfxsrxq", result)

    def test_five_params_hash_matches_shared_encoding(self):
        """Shared length-prefix expected (full hash): prrdp31y37ufw3lo7hly0oimjjvg_34m9ji30ocu4tw"""
        result = _compute_ext_cache_key({
            "key3": "value3", "key4": "value4",
            "key5": "value5", "key6": "value6", "key7": "value7",
        })
        self.assertEqual("prrdp31y37ufw3lo7hly0oimjjvg_34m9ji30ocu4tw", result)

    def test_order_independence_matches_shared_encoding(self):
        """Same keys in different insertion order must produce the same hash
        (mirrors TestCacheKeyComponentHashConsistency in Go)."""
        h1 = _compute_ext_cache_key({"key3": "value3", "key4": "value4",
                                      "key5": "value5", "key6": "value6", "key7": "value7"})
        h2 = _compute_ext_cache_key({"key7": "value7", "key4": "value4",
                                      "key6": "value6", "key5": "value5", "key3": "value3"})
        self.assertEqual(h1, h2)

    def test_at_cache_key_uses_atext_credential_type(self):
        """When ext_cache_key is present the credential type segment of the
        AT cache key must be 'atext' (not 'accesstoken'), matching Go/dotnet.

        Go:    {hid}-{env}-atext-{clientID}-{realm}-{scopes}-{hash}
        .NET:  {hid}-{env}-atext-{clientID}-{tenantId}-{scopes}-{hash}
        """
        cache = TokenCache()
        key_maker = cache.key_makers[TokenCache.CredentialType.ACCESS_TOKEN]
        key = key_maker(
            home_account_id="hid", environment="env", client_id="cid",
            realm="realm", target="scope",
            ext_cache_key="latlwkpewb_a0rcsmjvkecqt0_huumkw4sflzociike")
        self.assertEqual(
            "hid-env-atext-cid-realm-scope-latlwkpewb_a0rcsmjvkecqt0_huumkw4sflzociike",
            key)

    def test_at_cache_key_without_ext_uses_accesstoken(self):
        """Regular ATs (no ext_cache_key) must keep 'accesstoken' credential type."""
        cache = TokenCache()
        key_maker = cache.key_makers[TokenCache.CredentialType.ACCESS_TOKEN]
        key = key_maker(
            home_account_id="hid", environment="env", client_id="cid",
            realm="realm", target="scope")
        self.assertEqual("hid-env-accesstoken-cid-realm-scope", key)

    def test_full_at_cache_key(self):
        """Reproduce the full 'atext' cache key layout (mirrors MSAL .NET's
        CacheKeyExtensionTests expectedCacheKey1 shape) with the shared
        length-prefix hash for {key1:value1, key2:value2}.
        """
        cache = TokenCache()
        key_maker = cache.key_makers[TokenCache.CredentialType.ACCESS_TOKEN]
        ext_hash = _compute_ext_cache_key({"key1": "value1", "key2": "value2"})
        key = key_maker(
            home_account_id="",
            environment="login.windows.net",
            client_id="d3adb33f-c0de-ed0c-c0de-deadb33fc0d3",
            realm="common",
            target="r1/scope1 r1/scope2",
            ext_cache_key=ext_hash)
        expected = "-login.windows.net-atext-d3adb33f-c0de-ed0c-c0de-deadb33fc0d3-common-r1/scope1 r1/scope2-latlwkpewb_a0rcsmjvkecqt0_huumkw4sflzociike"
        self.assertEqual(expected, key)

    def test_second_full_at_cache_key(self):
        """Reproduce the 'atext' cache key layout (mirrors expectedCacheKey2 shape)
        with the shared length-prefix hash for {key3:value3, key4:value4}."""
        cache = TokenCache()
        key_maker = cache.key_makers[TokenCache.CredentialType.ACCESS_TOKEN]
        ext_hash = _compute_ext_cache_key({"key3": "value3", "key4": "value4"})
        key = key_maker(
            home_account_id="",
            environment="login.windows.net",
            client_id="d3adb33f-c0de-ed0c-c0de-deadb33fc0d3",
            realm="common",
            target="r1/scope1 r1/scope2",
            ext_cache_key=ext_hash)
        expected = "-login.windows.net-atext-d3adb33f-c0de-ed0c-c0de-deadb33fc0d3-common-r1/scope1 r1/scope2-jjoe9jgfmdtnj0rzuetsqy7kzs2m1xfnjjxwsfxsrxq"
        self.assertEqual(expected, key)

    def test_go_style_at_cache_key(self):
        """Reproduce the Go AccessToken.Key() *format* (segment layout):
        'testhid-env-atext-clientid-realm-user.read-{hash}'. The hash follows the
        shared length-prefix encoding (byte-identical to MSAL Go).
        """
        cache = TokenCache()
        key_maker = cache.key_makers[TokenCache.CredentialType.ACCESS_TOKEN]
        ext_hash = _compute_ext_cache_key({"key1": "value1", "key2": "value2"})
        key = key_maker(
            home_account_id="testhid",
            environment="env",
            client_id="clientid",
            realm="realm",
            target="user.read",
            ext_cache_key=ext_hash)
        expected = "testhid-env-atext-clientid-realm-user.read-latlwkpewb_a0rcsmjvkecqt0_huumkw4sflzociike"
        self.assertEqual(expected, key)


class TestExtCacheKeyCollisionResistance(unittest.TestCase):
    """The length-prefix ("netstring") serialization must be injective: no two
    semantically different component sets may serialize to the same pre-hash
    string. (The cache key itself is a SHA-256 digest, so the residual
    collision risk is only the cryptographically negligible hash-layer one;
    these tests pin the serialization layer that we control.)

    A plain ``key + value`` concatenation (the old scheme) is ambiguous and
    caused cache-slot collisions -- one FMI/agent-identity token entry could
    evict another, forcing redundant token re-fetches. These tests pin the
    boundary cases that the old scheme got wrong.
    """

    def test_key_value_boundary_ambiguity(self):
        # Old scheme: both -> "fmi_pathvalue".
        self.assertNotEqual(
            _compute_ext_cache_key({"fmi_path": "value"}),
            _compute_ext_cache_key({"fmi_pat": "hvalue"}))

    def test_multi_entry_boundary_ambiguity(self):
        # Old scheme: both -> "abcde".
        self.assertNotEqual(
            _compute_ext_cache_key({"a": "b", "cd": "e"}),
            _compute_ext_cache_key({"ab": "c", "d": "e"}))

    def test_value_containing_encoding_delimiters(self):
        # Values that themselves contain the "<len>:<data>" delimiters must not
        # be able to forge a different component layout.
        self.assertNotEqual(
            _compute_ext_cache_key({"a": "5:hello"}),
            _compute_ext_cache_key({"a5": "hello"}))
        self.assertNotEqual(
            _compute_ext_cache_key({"x": "1:y1:z"}),
            _compute_ext_cache_key({"x": "1:y", "1": "z"}))

    def test_injectivity_over_adversarial_alphabet(self):
        # Build many distinct component dicts from an adversarial alphabet that
        # stresses the length-prefix delimiters and UTF-8 boundaries, then assert
        # no two *distinct* dicts share a hash (and identical dicts agree).
        import itertools
        alphabet = ["", "0", "1", "9", ":", "|", "\\", "a", "ab",
                    u"\u00e9", u"\U0001F600", u"e\u0301"]
        seen = {}
        for k1, v1 in itertools.product(alphabet, repeat=2):
            for k2, v2 in itertools.product(alphabet, repeat=2):
                # Keys must be truthy to survive field-selection, and distinct so
                # the two pairs don't silently overwrite each other in a dict
                # (which would collapse intended 2-entry cases). Values must be
                # truthy too; empty values are dropped by field-selection.
                pairs = [(k, v) for k, v in ((k1, v1), (k2, v2)) if k and v]
                if len(pairs) == 2 and pairs[0][0] == pairs[1][0]:
                    continue
                comps = dict(pairs)
                if not comps:
                    continue
                # Canonicalize so logically-equal dicts compare equal.
                canonical = tuple(sorted(comps.items()))
                h = _compute_ext_cache_key(comps)
                if h in seen:
                    self.assertEqual(
                        seen[h], canonical,
                        "Hash collision between {!r} and {!r} -> {}".format(
                            dict(seen[h]), comps, h))
                else:
                    seen[h] = canonical
        # seen holds one entry per *distinct* component set, so its size proves
        # we exercised a meaningful number of unique inputs (not repeats).
        self.assertGreater(len(seen), 100)

    def test_utf8_byte_length_not_codepoint_length(self):
        # 'é' as one 2-byte codepoint (U+00E9) vs 'e' + combining acute accent
        # (U+0065 U+0301, 3 bytes) render alike but are distinct. A code-point
        # length prefix (len(str)) would risk a collision at some boundary; the
        # UTF-8 byte-length prefix keeps them apart and matches MSAL Go.
        precomposed = u"\u00e9"          # 1 code point, 2 UTF-8 bytes
        decomposed = u"e\u0301"          # 2 code points, 3 UTF-8 bytes
        self.assertNotEqual(
            _compute_ext_cache_key({"k": precomposed}),
            _compute_ext_cache_key({"k": decomposed}))
        # A concrete boundary pair the two length schemes disagree on:
        # code-point length would make these ambiguous, byte length does not.
        self.assertNotEqual(
            _compute_ext_cache_key({"k": u"\u00e9x"}),
            _compute_ext_cache_key({"k": u"e\u0301"}))

    def test_key_order_independence(self):
        self.assertEqual(
            _compute_ext_cache_key({"b": "2", "a": "1", "c": "3"}),
            _compute_ext_cache_key({"c": "3", "a": "1", "b": "2"}))

    def test_empty_and_single_entry_edges(self):
        self.assertEqual("", _compute_ext_cache_key(None))
        self.assertEqual("", _compute_ext_cache_key({}))
        self.assertEqual("", _compute_ext_cache_key({"fmi_path": ""}))
        single = _compute_ext_cache_key({"fmi_path": "p"})
        self.assertTrue(single)
        self.assertEqual(single, _compute_ext_cache_key({"fmi_path": "p"}))

    def test_golden_vectors(self):
        # Shared length-prefix golden vectors: byte-identical across the MSAL SDK
        # family (Go/.NET/Java/JS). Output is already lowercased.
        golden = {
            u"a0ry_zl4gccsdp7gnw927x8s0mrmnodv6tyilt0u07m":
                {"fmi_path": "agent-app-id"},
            u"cybgactkrvlzlen1aiwzwl3ay5krkyixommrobc-ri4":
                {"a": "b", "cd": "e"},
            u"n_lucewkadzv_nybtg-2wtorgf2nrns6ihlfa7vbuzg":
                {"fmi_path": "value"},
            u"tjtm16m-suk2_bkniblr25lyuki40qyceco7knuyu0k":
                {"fmi_pat": "hvalue"},
            u"xskzaoz4ibr3mznftyxctvg1ptuh-0fuzpty7ndbfls":
                {u"\u00e9": u"\u00e9"},
        }
        for expected, components in golden.items():
            self.assertEqual(expected, _compute_ext_cache_key(components))
